'''
   PyMFEM example 37 - Serial/Parallel Shared code
'''

import os
from os.path import expanduser, join

from numpy import sqrt, log, exp
import numpy as np
from numba import njit
from numba.types import float64

from mfem import mfem_mode

if mfem_mode == 'serial':
    import mfem.ser as mfem
    from mfem.ser import intArray, doubleArray
    use_parallel = False
else:
    import mfem.par as mfem
    from mfem.par import intArray, doubleArray
    use_parallel = True


@njit(float64(float64))
def inv_sigmoid(x):
    '''
    Inverse sigmoid function
    '''
    tol = 1e-12
    x = min(max(tol, x), 1.0-tol)
    return log(x/(1.0-x))


@njit(float64(float64))
def sigmoid(x):
    '''
    Sigmoid function
    '''
    if x >= 0:
        return 1.0/(1.0 + exp(-x))
    else:
        return exp(x)/(1.0 + exp(x))


@njit(float64(float64))
def der_sigmoid(x):
    '''
    Derivative of sigmoid function    
    '''
    tmp = sigmoid(-x)
    return tmp - tmp**2


def MappedGridFunctionCoefficient(gf, func):

    c_gf = mfem.GridFunctionCoefficient(gf)

    @mfem.jit.scalar(dependency=(c_gf,))
    def coeff(ptx, c_gf):
        return func(c_gf)
    return coeff


def DiffMappedGridFunctionCoefficient(gf, other_gf, func, comp=1):
    c_gf = mfem.GridFunctionCoefficient(gf)
    c_ogf = mfem.GridFunctionCoefficient(other_gf)

    @mfem.jit.scalar(dependency=(c_gf, c_ogf))
    def coeff(ptx, c_gf, c_ogf):
        return func(c_gf) - func(c_ogf)
    return coeff


class SIMPInterpolationCoefficient():
    '''
    Python Note: 
       Approach here is to replace Eval in C++ example using the dependency 
       feature of mfem.jit.

       In order to avoid repeating Numba-Jitting in iteration loop, we use
       SetGridFunction to update the GridFunction referred from 
       GridFunctionCoefficient.
    '''

    def __init__(self, rho_filter, min_val=1e-6, max_val=1.0, exponent=3):
        val = mfem.GridFunctionCoefficient(rho_filter)

        @mfem.jit.scalar(dependency=(val,))
        def coeff(ptx, val):
            coeff = min_val + val**exponent*(max_val-min_val)
            return coeff

        self.c_gf = val
        self.coeff = coeff

    def Update(self, rho_filter):
        self.c_gf.SetGridFunction(rho_filter)


def VolumeForceCoefficient(r, center, force):

    @mfem.jit.vector(shape=(len(center),))
    def coeff(ptx):
        cr = sqrt(sum((ptx - center)**2))
        if cr < r:
            return np.array((force[0], force[1]))
        else:
            return np.array((0.0, 0.0))
    return coeff



import sys
import numpy as np
import gc
from scipy.sparse import csr_matrix

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True

    from mfem.common.mpi_debug import nicePrint as print

    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank

else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def test():
    dim = 3
    max_attr = 5
    sigma_attr_coefs = mfem.MatrixCoefficientArray(max_attr)
    sigma_attr = mfem.intArray(max_attr)


    tensors =  [mfem.DenseMatrix(np.ones((3,3))*i) for i in range(max_attr)]
    tensor_coefs =  [mfem.MatrixConstantCoefficient(mat) for mat in tensors]
    sigma_array = mfem.MatrixCoefficientArray(tensor_coefs)
    sigma_attr = mfem.intArray([1,2,3,4,5])
    sigmaCoef = mfem.PWMatrixCoefficient(dim, sigma_attr, sigma_array, False)

    for ti, tensor in enumerate(tensors): 
        # add matrix coefficient to list
        xx = mfem.MatrixConstantCoefficient(tensor)
        sigma_attr_coefs[ti] = xx
        sigma_attr[ti] = ti+1   

    # Create PW Matrix Coefficient
    sigmaCoef = mfem.PWMatrixCoefficient(dim, sigma_attr, sigma_attr_coefs, False)

    sigmaCoef = mfem.PWMatrixCoefficient(dim, sigma_attr, tensor_coefs, False)
    tensor_coefs = mfem.MatrixCoefficientArray([mfem.MatrixConstantCoefficient(mat) for mat in tensors])
    sigmaCoef = mfem.PWMatrixCoefficient(dim, sigma_attr, tensor_coefs, False)

    data = tensor_coefs.GetData()
    tensor_coefs2 = mfem.MatrixCoefficientArray(data, 5, False)
    sigmaCoef = mfem.PWMatrixCoefficient(dim, sigma_attr, tensor_coefs2, False)

    print("exiting")


if __name__ == '__main__':
    import tracemalloc

    tracemalloc.start()

    print(tracemalloc.get_traced_memory())

    for i in range(3000):
        test()
        print(tracemalloc.get_traced_memory())
    snapshot = tracemalloc.take_snapshot()
    top_stats = snapshot.statistics('lineno')
    for stat in top_stats[:10]: 
        print(stat)




import warnings
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem   
else:
    import mfem.ser as mfem

def make_matrix(x, y, z):
    pass

def do_findpoints(mesh, *args):
    sdim = mesh.SpaceDimension()
    
    shape = args[0].shape
    size= len(args[0].flatten())
    ptx = np.vstack([t.flatten() for t in args]).transpose().flatten()
    
    ptx2 = mfem.Vector(ptx)
    point_mat = mfem.DenseMatrix(ptx2.GetData(), size, sdim)
    elem_id = mfem.intArray()
    ips = mfem.IntegrationPointArray()
        
    num_found = mesh.FindPoints(point_mat, elem_id, ips, True)
    elem_id = np.array(elem_id.ToList())
    return v, elem_id, ips

def eval_at_points(gf, *args):
    args = [np.atleast_1d(np.asarray(t)) for t in args]
    mesh = gf.FESpace().Mesh()
    v, elem_id, ips = findpoints(mesh, *args)
    
def findpoints(mesh, *args):
    '''
    *args : x, y or x, y, z
    '''
    sdim = mesh.SpaceDimension()
    
    if len(args) != 3 and sdim == 3:
        assert False, "SpaceDimension = 3, pass x, y, z"
    elif len(args) != 2 and sdim == 2:
        assert False, "SpaceDimension = 3, pass x, y"
    elif len(args) != 1 and sdim == 1:
        assert False, "SpaceDimension = 3, pass x"
    else:
        args = [np.atleast_1d(np.array(t)) for t in args]
        return do_findpoints(mesh, *args)




'''
   MFEM example 35p
      See c++ version in the MFEM library for more detail 

   Sample runs:  mpirun -np 4 python ex35p.py -p 0 -o 2
                 mpirun -np 4 python ex35p.py -p 0 -o 2 -pbc '22 23 24' -em 0
                 mpirun -np 4 python ex35p.py -p 1 -o 1 -rp 2
                 mpirun -np 4 python ex35p.py -p 1 -o 2
                 mpirun -np 4 python ex35p.py -p 2 -o 1 -rp 2 -c 15
'''
import os
import sys
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)

mixed = True
visualization = True
pa = False
slu_solver = False  # SuperLU option not supported


def run(mesh_file="",
        order=1,
        ser_ref_levels=0,
        par_ref_levels=0,
        prob=0,
        mode=1,
        herm_conv=True,
        epsilon=1.0,
        sigma=2.,
        mu=1.0,
        device='cpu',
        port_bc_attr=None,
        freq=-1.0,):

    mesh_file = "../data/fichera-mixed.mesh"
    if not mixed or pa:
        mesh_file = "../data/fichera.mesh"

    omega = 2*pi
    if freq > 0.0:
        omega = 2.0 * pi * freq

    if (len(port_bc_attr) == 0 and
        (mesh_file == "../data/fichera-mixed.mesh" or
         mesh_file == "../data/fichera.mesh")):
        port_bc_attr = mfem.intArray([7, 8, 11, 12])

    port_bc_attr.Print()
    conv = mfem.ComplexOperator.HERMITIAN if herm_conv else mfem.ComplexOperator.BLOCK_SYMMETRIC
    # 3. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device)
    if myid == 0:
        device.Print()

    # 4. Read the mesh from the given mesh file. We can handle triangular,
    #    quadrilateral, tetrahedral, hexahedral, surface and volume meshes
    #    with the same code.
    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()

    # 5. Refine the serial mesh on all processors to increase the resolution.
    for i in range(ser_ref_levels):
        mesh.UniformRefinement()

    # 6a. Define a parallel mesh by a partitioning of the serial mesh. Refine
    #    this mesh further in parallel to increase the resolution. Once the
    #    parallel mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    for i in range(par_ref_levels):
        pmesh.UniformRefinement()

    # 6b. Extract a submesh covering a portion of the boundary
    pmesh_port = mfem.ParSubMesh.CreateFromBoundary(pmesh, port_bc_attr)

    # 7a. Define a parallel finite element space on the parallel mesh. Here we
    #     use continuous Lagrange, Nedelec, or Raviart-Thomas finite elements
    #     of the specified order.
    if dim == 1 and prob != 0:
        if myid == 0:
            print(
                "Switching to problem type 0, H1 basis functions, for 1 dimensional mesh.")
        prob = 0

    if prob == 0:
        fec = mfem.H1_FECollection(order, dim)
    elif prob == 1:
        fec = mfem.ND_FECollection(order, dim)
    elif prob == 2:
        fec = mfem.RT_FECollection(order-1, dim)
    else:
        assert False, "unknown problem"
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    size = fespace.GlobalTrueVSize()
    if myid == 0:
        print("Number of finite element unknowns: " + str(size))

    # 7b. Define a parallel finite element space on the sub-mesh. Here we
    #     use continuous Lagrange, Nedelec, or L2 finite elements of
    #     the specified order.
    if prob == 0:
        fec_port = mfem.H1_FECollection(order, dim-1)
    elif prob == 1:
        if dim == 3:
            fec_port = mfem.ND_FECollection(order, dim-1)
        else:
            fec_port = mfem.L2_FECollection(order - 1, dim-1,
                                            mfem.BasisType.GaussLegendre,
                                            mfem.FiniteElement.INTEGRAL)
    elif prob == 2:
        fec_port = mfem.L2_FECollection(order - 1, dim-1,
                                        mfem.BasisType.GaussLegendre,
                                        mfem.FiniteElement.INTEGRAL)
    else:
        assert False, "should not reach here"

    fespace_port = mfem.ParFiniteElementSpace(pmesh_port, fec_port)
    size_port = fespace_port.GlobalTrueVSize()
    if myid == 0:
        print("Number of finite element port BC unknowns: " + str(size_port))

    # 8a. Define a parallel grid function on the SubMesh which will contain
    #     the field to be applied as a port boundary condition.
    port_bc = mfem.ParGridFunction(fespace_port)
    port_bc.Assign(0.0)

    SetPortBC(prob, dim, mode, port_bc)

    # 8b. Save the SubMesh and associated port boundary condition in parallel.
    #     This output can be viewed later using GLVis:
    #     "glvis -np <np> -m port_mesh -g port_mode"
    pmesh_port.Print("port_mesh"+smyid, 8)
    port_bc.Save("port_mode"+smyid, 8)

    # 8c. Send the port bc, computed on the SubMesh, to a GLVis server.
    if visualization and dim == 3:
        port_sock = mfem.socketstream("localhost", 19916)
        port_sock << "parallel " << num_procs << " " << myid << "\n"
        port_sock.precision(8)
        port_sock << "solution\n" << pmesh_port << port_bc
        port_sock << "window_title 'Port BC'"
        port_sock << "window_geometry 0 0 400 350"
        port_sock.flush()

    # 9. Determine the list of true (i.e. parallel conforming) essential
    #    boundary dofs. In this example, the boundary conditions are defined
    #    using an eigenmode of the appropriate type computed on the SubMesh.
    ess_tdof_list = mfem.intArray()
    if pmesh.bdr_attributes.Size():
        ess_bdr = mfem.intArray([1]*pmesh.bdr_attributes.Max())
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 10. Set up the parallel linear form b(.) which corresponds to the
    #     right-hand side of the FEM linear system.
    b = mfem.ParComplexLinearForm(fespace, conv)
    b.Assign(0.0)

    # 11a. Define the solution vector u as a parallel complex finite element
    #      grid function corresponding to fespace. Initialize u to equal zero.
    u = mfem.ParComplexGridFunction(fespace)
    u.Assign(0.0)
    pmesh_port.Transfer(port_bc, u.real())

    # 11b. Send the transferred port bc field to a GLVis server.

    full_bc = mfem.ParGridFunction(fespace)
    port_to_full = mfem.ParTransferMap(port_bc, full_bc)

    full_bc.Assign(0.0)
    port_to_full.Transfer(port_bc, full_bc)

    if visualization:
        full_sock = mfem.socketstream("localhost", 19916)
        full_sock << "parallel " << num_procs << " " << myid << "\n"
        full_sock.precision(8)
        full_sock << "solution\n" << pmesh << full_bc
        full_sock << "window_title 'Transferred BC'"
        full_sock << "window_geometry 400 0 400 350"
        full_sock.flush()

    # 12. Set up the parallel sesquilinear form a(.,.) on the finite element
    #     space corresponding to the damped harmonic oscillator operator of the
    #     appropriate type:
    #
    #     0) A scalar H1 field
    #        -Div(a Grad) - omega^2 b + i omega c
    #
    #     1) A vector H(Curl) field
    #        Curl(a Curl) - omega^2 b + i omega c
    #
    #     2) A vector H(Div) field
    #        -Grad(a Div) - omega^2 b + i omega c
    #
    stiffnessCoef = mfem.ConstantCoefficient(1.0/mu)
    massCoef = mfem.ConstantCoefficient(-omega * omega * epsilon)
    lossCoef = mfem.ConstantCoefficient(omega * sigma)
    negMassCoef = mfem.ConstantCoefficient(omega * omega * epsilon)

    a = mfem.ParSesquilinearForm(fespace, conv)

    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    if prob == 0:
        a.AddDomainIntegrator(mfem.DiffusionIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.MassIntegrator(massCoef),
                              mfem.MassIntegrator(lossCoef))
    elif prob == 1:
        a.AddDomainIntegrator(mfem.CurlCurlIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef),
                              mfem.VectorFEMassIntegrator(lossCoef))
    elif prob == 2:
        a.AddDomainIntegrator(mfem.DivDivIntegrator(stiffnessCoef),
                              None)
        a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef),
                              mfem.VectorFEMassIntegrator(lossCoef))
    else:
        assert False, "Unknown probm"

    # 13. Assemble the parallel bilinear form and the corresponding linear
    #     system, applying any necessary transformations such as: parallel
    #     assembly, eliminating boundary conditions, applying conforming
    #     constraints for non-conforming AMR, etc.
    a.Assemble()
    A = mfem.OperatorHandle()
    B = mfem.Vector()
    U = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, u, b, A, U, B)

    if myid == 0:
        print("Size of linear system: " + str(2*size))

    if not slu_solver:
        # 14a. Set up the parallel bilinear form for the preconditioner
        #      corresponding to the appropriate operator
        #
        #      0) A scalar H1 field
        #         -Div(a Grad) - omega^2 b + i omega c
        #
        #      1) A vector H(Curl) field
        #         Curl(a Curl) + omega^2 b + i omega c
        #
        #      2) A vector H(Div) field
        #         -Grad(a Div) - omega^2 b + i omega c
        pcOp = mfem.ParBilinearForm(fespace)
        if pa:
            pcOp.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)

        if prob == 0:
            pcOp.AddDomainIntegrator(mfem.DiffusionIntegrator(stiffnessCoef))
            pcOp.AddDomainIntegrator(mfem.MassIntegrator(massCoef))
            pcOp.AddDomainIntegrator(mfem.MassIntegrator(lossCoef))
        elif prob == 1:
            pcOp.AddDomainIntegrator(mfem.CurlCurlIntegrator(stiffnessCoef))
            pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(negMassCoef))
            pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(lossCoef))
        elif prob == 2:
            pcOp.AddDomainIntegrator(mfem.DivDivIntegrator(stiffnessCoef))
            pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(massCoef))
            pcOp.AddDomainIntegrator(mfem.VectorFEMassIntegrator(lossCoef))
        pcOp.Assemble()

        # 14b. Define and apply a parallel FGMRES solver for AU=B with a block
        #      diagonal preconditioner based on the appropriate multigrid
        #      preconditioner from hypre.
        blockTrueOffsets = mfem.intArray()
        blockTrueOffsets.SetSize(3)
        blockTrueOffsets[0] = 0
        blockTrueOffsets[1] = A.Height() // 2
        blockTrueOffsets[2] = A.Height() // 2
        blockTrueOffsets.PartialSum()

        BDP = mfem.BlockDiagonalPreconditioner(blockTrueOffsets)

        if pa:
            pc_r = mfem.OperatorJacobiSmoother(pcOp, ess_tdof_list)
            pc_i = None
        else:
            PCOp = mfem.OperatorHandle()
            pcOp.FormSystemMatrix(ess_tdof_list, PCOp)
            if prob == 0:
                pc_r = mfem.HypreBoomerAMG(PCOp.AsHypreParMatrix())
            elif prob == 1:
                pc_r = mfem.HypreAMS(PCOp.AsHypreParMatrix(), fespace)
            elif prob == 2:
                if dim == 2:
                    pc_r = mfem.HypreAMS(PCOp.AsHypreParMatrix(), fespace)
                else:
                    pc_r = mfem.HypreADS(PCOp.AsHypreParMatrix(), fespace)

        pc_i = mfem.ScaledOperator(pc_r,
                                   -1 if conv == mfem.ComplexOperator.HERMITIAN else 1)
        BDP.SetDiagonalBlock(0, pc_r)
        BDP.SetDiagonalBlock(1, pc_i)

        fgmres = mfem.FGMRESSolver(MPI.COMM_WORLD)
        fgmres.SetPreconditioner(BDP)
        fgmres.SetOperator(A.Ptr())
        fgmres.SetRelTol(1e-6)
        fgmres.SetMaxIter(1000)
        fgmres.SetPrintLevel(1)
        fgmres.Mult(B, U)

    # 15. Recover the parallel grid function corresponding to U. This is the
    #     local finite element solution on each processor.
    a.RecoverFEMSolution(U, b, u)

    # 16. Save the refined mesh and the solution in parallel. This output can be
    #     viewed later using GLVis: "glvis -np <np> -m mesh -g sol_r" or
    #     "glvis -np <np> -m mesh -g sol_i".
    pmesh.Print("mesh"+smyid, 8)
    u.real().Save("sol_r"+smyid, 8)
    u.imag().Save("sol_i"+smyid, 8)

    # 17. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock_r = mfem.socketstream("localhost", 19916)
        sol_sock_r << "parallel " << num_procs << " " << myid << "\n"
        sol_sock_r.precision(8)
        sol_sock_r << "solution\n" << pmesh << u.real(
        ) << "window_title 'Solution: Real Part'"
        sol_sock_r << "window_geometry 800 0 400 350"
        sol_sock_r.flush()

        sol_sock_i = mfem.socketstream("localhost", 19916)
        sol_sock_i << "parallel " << num_procs << " " << myid << "\n"
        sol_sock_i.precision(8)
        sol_sock_i << "solution\n" << pmesh << u.imag(
        ) << "window_title 'Solution: Imag Part'"
        sol_sock_i << "window_geometry 1200 0 400 350"
        sol_sock_i.flush()

    if visualization:
        u_t = mfem.ParGridFunction(fespace)
        u_t.Assign(u.real())
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock.precision(8)
        sol_sock << "solution\n" << pmesh << u_t
        sol_sock << "window_title 'Harmonic Solution (t = 0.0 T)'"
        sol_sock << "window_geometry 0 432 600 450"
        sol_sock << "pause\n"
        sol_sock.flush()

        if myid == 0:
            print(
                "GLVis visualization paused. Press space (in the GLVis window) to resume it.")
        num_frames = 32
        i = 0

        # Let's plot one wave cycle...
        for i in range(160):
            t = (i % num_frames) / num_frames
            oss = "Harmonic Solution (t = " + str(t) + " T)"
            dd = (cos(2.0 * pi * t)*u.real().GetDataArray() +
                  sin(2.0 * pi * t)*u.imag().GetDataArray())
            # we can not load numpy directly...(sorry)
            u_t.Assign(mfem.Vector(dd))
            sol_sock << "parallel " << num_procs << " " << myid << "\n"
            sol_sock << "solution\n" << pmesh << u_t
            sol_sock << "window_title '" << oss << "'"
            sol_sock.flush()


'''
   Solves the eigenvalue problem -Div(Grad x) = lambda x with homogeneous
   Dirichlet boundary conditions on the boundary of the domain. Returns mode
   number "mode" (counting from zero) in the ParGridFunction "x".
'''


def ScalarWaveGuide(mode, x):
    nev = max(mode + 2, 5)
    seed = 75

    fespace = x.ParFESpace()
    pmesh = fespace.GetParMesh()

    if pmesh.bdr_attributes.Size() > 0:
        ess_bdr = mfem.intArray([1]*pmesh.bdr_attributes.Max())
    else:
        ess_bdr = mfem.intArray()

    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator())
    a.Assemble()
    a.EliminateEssentialBCDiag(ess_bdr, 1.0)
    a.Finalize()

    m = mfem.ParBilinearForm(fespace)
    m.AddDomainIntegrator(mfem.MassIntegrator())
    m.Assemble()

    # shift the eigenvalue corresponding to eliminated dofs to a large value
    m.EliminateEssentialBCDiag(ess_bdr, sys.float_info.min)
    m.Finalize()

    A = a.ParallelAssemble()
    M = m.ParallelAssemble()

    amg = mfem.HypreBoomerAMG(A)
    amg.SetPrintLevel(0)

    lobpcg = mfem.HypreLOBPCG(MPI.COMM_WORLD)
    lobpcg.SetNumModes(nev)
    lobpcg.SetRandomSeed(seed)
    lobpcg.SetPreconditioner(amg)
    lobpcg.SetMaxIter(200)
    lobpcg.SetTol(1e-8)
    lobpcg.SetPrecondUsageMode(1)
    lobpcg.SetPrintLevel(1)
    lobpcg.SetMassMatrix(M)
    lobpcg.SetOperator(A)
    lobpcg.Solve()

    x.Assign(lobpcg.GetEigenvector(mode))


'''
   Solves the eigenvalue problem -Curl(Curl x) = lambda x with homogeneous
   Dirichlet boundary conditions, on the tangential component of x, on the
   boundary of the domain. Returns mode number "mode" (counting from zero) in
   the ParGridFunction "x".
'''


def VectorWaveGuide(mode, x):
    nev = max(mode + 2, 5)
    fespace = x.ParFESpace()
    pmesh = fespace.GetParMesh()

    if pmesh.bdr_attributes.Size() > 0:
        ess_bdr = mfem.intArray([1]*pmesh.bdr_attributes.Max())
    else:
        ess_bdr = mfem.intArray()

    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator())
    a.Assemble()
    a.EliminateEssentialBCDiag(ess_bdr, 1.0)
    a.Finalize()

    m = mfem.ParBilinearForm(fespace)
    m.AddDomainIntegrator(mfem.VectorFEMassIntegrator())
    m.Assemble()

    # shift the eigenvalue corresponding to eliminated dofs to a large value
    m.EliminateEssentialBCDiag(ess_bdr, sys.float_info.min)
    m.Finalize()

    A = a.ParallelAssemble()
    M = m.ParallelAssemble()

    ams = mfem.HypreAMS(A, fespace)
    ams.SetPrintLevel(0)
    ams.SetSingularProblem()

    ame = mfem.HypreAME(MPI.COMM_WORLD)
    ame.SetNumModes(nev)
    ame.SetPreconditioner(ams)
    ame.SetMaxIter(100)
    ame.SetTol(1e-8)
    ame.SetPrintLevel(1)
    ame.SetMassMatrix(M)
    ame.SetOperator(A)
    ame.Solve()

    x.Assign(ame.GetEigenvector(mode))


'''
   Solves the eigenvalue problem -Div(Grad x) = lambda x with homogeneous
   Neumann boundary conditions on the boundary of the domain. Returns mode
   number "mode" (counting from zero) in the ParGridFunction "x_l2". Note that
   mode 0 is a constant field so higher mode numbers are often more
   interesting. The eigenmode is solved using continuous H1 basis of the
   appropriate order and then projected onto the L2 basis and returned.
'''


def PseudoScalarWaveGuide(mode, x_l2):
    nev = max(mode + 2, 5)
    seed = 75

    fespace_l2 = x_l2.ParFESpace()
    pmesh = fespace_l2.GetParMesh()
    order_l2 = fespace_l2.FEColl().GetOrder()

    fec = mfem.H1_FECollection(order_l2+1, pmesh.Dimension())
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    x = mfem.ParGridFunction(fespace)
    x.Assign(0.0)

    xCoef = mfem.GridFunctionCoefficient(x)
    if mode == 0:
        x.Assign(1.0)
        x_l2.ProjectCoefficient(xCoef)

    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator())
    a.AddDomainIntegrator(mfem.MassIntegrator())
    a.Assemble()
    a.Finalize()

    m = mfem.ParBilinearForm(fespace)
    m.AddDomainIntegrator(mfem.MassIntegrator())
    m.Assemble()
    m.Finalize()

    A = a.ParallelAssemble()
    M = m.ParallelAssemble()

    amg = mfem.HypreBoomerAMG(A)
    amg.SetPrintLevel(0)

    lobpcg = mfem.HypreLOBPCG(MPI.COMM_WORLD)
    lobpcg.SetNumModes(nev)
    lobpcg.SetRandomSeed(seed)
    lobpcg.SetPreconditioner(amg)
    lobpcg.SetMaxIter(200)
    lobpcg.SetTol(1e-8)
    lobpcg.SetPrecondUsageMode(1)
    lobpcg.SetPrintLevel(1)
    lobpcg.SetMassMatrix(M)
    lobpcg.SetOperator(A)
    lobpcg.Solve()

    x.Assign(lobpcg.GetEigenvector(mode))
    x_l2.ProjectCoefficient(xCoef)

# Compute eigenmode "mode" of either a Dirichlet or Neumann Laplacian or of a
# Dirichlet curl curl operator based on the problem type and dimension of the
# domain.


def SetPortBC(prob, dim,  mode, port_bc):
    if prob == 0:
        ScalarWaveGuide(mode, port_bc)
    elif prob == 1:
        if dim == 3:
            VectorWaveGuide(mode, port_bc)
        else:
            PseudoScalarWaveGuide(mode, port_bc)
    elif prob == 2:
        PseudoScalarWaveGuide(mode, port_bc)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex35p (Port Boundary Conditions using SubMesh Transfers)')
    parser.add_argument('-m', '--mesh',
                        default='inline-quad.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument("-rs", "--refine-serial",
                        action='store', type=int, default=1,
                        help="Number of times to refine the mesh uniformly in serial.")
    parser.add_argument("-rp", "--refine-parallel",
                        action='store', type=int, default=1,
                        help="Number of times to refine the mesh uniformly in paralle.")
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-p", "--problem-type",
                        action='store', type=int, default=0,
                        help="\n".join(["Choose between 0: H_1, 1: H(Curl), or 2: H(Div) "
                                        "damped harmonic oscillator."]))
    parser.add_argument("-em", "--eigenmode",
                        action='store', type=int, default=1,
                        help="Choose the index of the port eigenmode.")
    parser.add_argument("-a", "--stiffness-coef",
                        action='store', type=float, default=0.0,
                        help="Stiffness coefficient (spring constant or 1/mu).")
    parser.add_argument("-b", "--mass-coef",
                        action='store', type=float, default=1.0,
                        help="Mass coefficient (or epsilon).")
    parser.add_argument("-c", "--damping-coef",
                        action='store', type=float, default=2.0,
                        help="Damping coefficient (or sigma).")
    parser.add_argument("-mu", "--permeability",
                        action='store', type=float, default=1.0,
                        help="Permeability of free space (or 1/(spring constant)).")
    parser.add_argument("-eps", "--permittivity",
                        action='store', type=float, default=1.0,
                        help="Permittivity of free space (or mass constant).")
    parser.add_argument("-sigma", "--conductivity",
                        action='store', type=float, default=2.0,
                        help="Conductivity (or damping constant).")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=-1.0,
                        help="Set the frequency for the exact")
    parser.add_argument("-pbc", "--port-bc-attr",
                        action='store', type=str, default="",
                        help="Attributes of port boundary condition")
    parser.add_argument("-herm", "--hermitian",
                        action='store_true',
                        default=True,
                        help="Do not use convention for Hermitian operators.")
    parser.add_argument("-no-herm", "--no-hermitian",
                        action='store_true',
                        default=False,
                        help="Do not use convention for Hermitian operators.")
    parser.add_argument("-no-vis", "--no-visualization",
                        action='store_true', default=False,
                        help='Enable GLVis visualization')
    parser.add_argument("-hex", "--hex-mesh",
                        action='store_true',  default=False,
                        help="Mixed mesh of hexahedral mesh.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")

    args = parser.parse_args()

    herm = False if args.no_hermitian else True
    args.hermitian = herm
    args.no_hermitian = not herm

    if myid == 0:
        parser.print_options(args)

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    port_bc_attr = [int(x) for x in args.port_bc_attr.split(' ') if len(x) > 0]
    globals()["mixed"] = not args.hex_mesh
    globals()["pa"] = args.partial_assembly
    globals()["visualization"] = not args.no_visualization

    if args.stiffness_coef != 0.0:
        mu = 1.0/args.stiffness_coef
    elif args.permeability != 1.0:
        mu = args.permeability
    else:
        mu = 1.0

    if args.permittivity != 1.0:
        epsilon = args.permittivity
    elif args.mass_coef != 1.0:
        epsilon = args.mass_coef
    else:
        epsilon = 1.0

    if args.conductivity != 2.0:
        sigma = args.conductivity
    elif args.damping_coef != 2.0:
        sigma = args.damping_coef
    else:
        sigma = 2.0

    run(mesh_file=meshfile,
        order=args.order,
        ser_ref_levels=args.refine_serial,
        par_ref_levels=args.refine_parallel,
        prob=args.problem_type,
        mode=args.eigenmode,
        herm_conv=herm,
        epsilon=epsilon,
        sigma=sigma,
        mu=mu,
        device=args.device,
        port_bc_attr=port_bc_attr,
        freq=args.frequency)



'''
   MFEM example 9
      This is a version of Example 1 with a simple adaptive mesh
      refinement loop. 
      See c++ version in the MFEM library for more detail 
'''
import os
from mfem import path
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname, exists
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc


# Inflow boundary condition (zero for the problems considered in this example)
class inflow_coeff(mfem.PyCoefficient):
    def EvalValue(self, x):
        return 0


class FE_Evolution(mfem.PyTimeDependentOperator):
    def __init__(self, M, K, b):
        mfem.PyTimeDependentOperator.__init__(self, M.Size())

        self.K = K
        self.M = M
        self.b = b
        self.z = mfem.Vector(M.Size())
        self.zp = np.zeros(M.Size())
        self.M_prec = mfem.DSmoother()
        self.M_solver = mfem.CGSolver()
        self.M_solver.SetPreconditioner(self.M_prec)
        self.M_solver.SetOperator(M)
        self.M_solver.iterative_mode = False
        self.M_solver.SetRelTol(1e-9)
        self.M_solver.SetAbsTol(0.0)
        self.M_solver.SetMaxIter(100)
        self.M_solver.SetPrintLevel(0)


#    def EvalMult(self, x):
#        if you want to impolement Mult in using python objects,
#        such as numpy.. this needs to be implemented and don't
#        overwrite Mult

    def Mult(self, x, y):
        self.K.Mult(x, self.z)
        self.z += self.b
        self.M_solver.Mult(self.z, y)


def run(ref_levels=2,
        order=3,
        ode_solver_type=4,
        t_final=10,
        dt=0.01,
        problem=0,
        vis_steps=5,
        meshfile='',
        visualization=False,
        device='cpu',
        pa=False,
        paraview=False):

    device = mfem.Device(device)
    device.Print()

    # 2. Read the mesh from the given mesh file. We can handle geometrically
    #    periodic meshes in this code.

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()

    # 3. Define the ODE solver used for time integration. Several explicit
    #    Runge-Kutta methods are available.
    ode_solver = None
    if ode_solver_type == 1:
        ode_solver = mfem.ForwardEulerSolver()
    elif ode_solver_type == 2:
        ode_solver = mfem.RK2Solver(1.0)
    elif ode_solver_type == 3:
        ode_solver = mfem.RK3SSolver()
    elif ode_solver_type == 4:
        ode_solver = mfem.RK4Solver()
    elif ode_solver_type == 6:
        ode_solver = mfem.RK6Solver()
    else:
        print("Unknown ODE solver type: " + str(ode_solver_type))
        exit

    # 4. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement, where 'ref_levels' is a
    #    command-line parameter. If the mesh is of NURBS type, we convert it to
    #    a (piecewise-polynomial) high-order mesh.
    for lev in range(ref_levels):
        mesh.UniformRefinement()
        if mesh.NURBSext:
            mesh.SetCurvature(max(order, 1))
    bb_min, bb_max = mesh.GetBoundingBox(max(order, 1))

    # 5. Define the discontinuous DG finite element space of the given
    #    polynomial order on the refined mesh.
    fec = mfem.DG_FECollection(order, dim, mfem.BasisType.GaussLobatto)
    fes = mfem.FiniteElementSpace(mesh, fec)

    print("Number of unknowns: " + str(fes.GetVSize()))

    #
    #  Define coefficient using VecotrPyCoefficient and PyCoefficient
    #  A user needs to define EvalValue method
    #
    class velocity_coeff(mfem.VectorPyCoefficient):
        def EvalValue(self, x):
            dim = len(x)

            center = (bb_min + bb_max)/2.0
            # map to the reference [-1,1] domain
            X = 2 * (x - center) / (bb_max - bb_min)
            if problem == 0:
                if dim == 1:
                    v = [1.0, ]
                elif dim == 2:
                    v = [sqrt(2./3.), sqrt(1./3)]
                elif dim == 3:
                    v = [sqrt(3./6.), sqrt(2./6), sqrt(1./6.)]
            elif (problem == 1 or problem == 2):
                # Clockwise rotation in 2D around the origin
                w = pi/2
                if dim == 1:
                    v = [1.0, ]
                elif dim == 2:
                    v = [w*X[1],  - w*X[0]]
                elif dim == 3:
                    v = [w*X[1],  - w*X[0],  0]
            elif (problem == 3):
                # Clockwise twisting rotation in 2D around the origin
                w = pi/2
                d = max((X[0]+1.)*(1.-X[0]), 0.) * max((X[1]+1.)*(1.-X[1]), 0.)
                d = d ** 2
                if dim == 1:
                    v = [1.0, ]
                elif dim == 2:
                    v = [d*w*X[1],  - d*w*X[0]]
                elif dim == 3:
                    v = [d*w*X[1],  - d*w*X[0],  0]
            return v

    class u0_coeff(mfem.PyCoefficient):
        def EvalValue(self, x):
            dim = len(x)

            center = (bb_min + bb_max)/2.0
            # map to the reference [-1,1] domain
            X = 2 * (x - center) / (bb_max - bb_min)
            if (problem == 0 or problem == 1):
                if dim == 1:
                    return exp(-40. * (X[0]-0.5)**2)
                elif (dim == 2 or dim == 3):
                    rx = 0.45
                    ry = 0.25
                    cx = 0.
                    cy = -0.2
                    w = 10.
                    if dim == 3:
                        s = (1. + 0.25*cos(2 * pi * x[2]))
                        rx = rx * s
                        ry = ry * s
                    return (erfc(w * (X[0]-cx-rx)) * erfc(-w*(X[0]-cx+rx)) *
                            erfc(w * (X[1]-cy-ry)) * erfc(-w*(X[1]-cy+ry)))/16.

            elif problem == 2:
                rho = hypot(x[0], x[1])
                phi = arctan2(x[1], x[0])
                return (sin(pi * rho) ** 2) * sin(3*phi)
            elif problem == 3:
                return sin(pi * X[0]) * sin(pi * X[1])

            return 0.0

    # 6. Set up and assemble the bilinear and linear forms corresponding to the
    #    DG discretization. The DGTraceIntegrator involves integrals over mesh
    #    interior faces.

    velocity = velocity_coeff(dim)
    inflow = inflow_coeff()
    u0 = u0_coeff()

    m = mfem.BilinearForm(fes)
    m.AddDomainIntegrator(mfem.MassIntegrator())
    k = mfem.BilinearForm(fes)
    k.AddDomainIntegrator(mfem.ConvectionIntegrator(velocity, -1.0))
    k.AddInteriorFaceIntegrator(
        mfem.TransposeIntegrator(mfem.DGTraceIntegrator(velocity, 1.0, -0.5)))
    k.AddBdrFaceIntegrator(
        mfem.TransposeIntegrator(mfem.DGTraceIntegrator(velocity, 1.0, -0.5)))

    b = mfem.LinearForm(fes)
    b.AddBdrFaceIntegrator(
        mfem.BoundaryFlowIntegrator(inflow, velocity, -1.0, -0.5))

    m.Assemble()
    m.Finalize()
    skip_zeros = 0
    k.Assemble(skip_zeros)
    k.Finalize(skip_zeros)
    b.Assemble()

    # 7. Define the initial conditions, save the corresponding grid function to
    #    a file
    u = mfem.GridFunction(fes)
    u.ProjectCoefficient(u0)

    mesh.Print('ex9.mesh', 8)
    u.Save('ex9-init.gf', 8)

    if paraview:
        pd = mfem.ParaViewDataCollection("Example9", mesh)
        pd.SetPrefixPath("ParaView")
        pd.RegisterField("solution", u)
        pd.SetLevelsOfDetail(order)
        pd.SetDataFormat(mfem.VTKFormat_BINARY)
        pd.SetHighOrderOutput(True)
        pd.SetCycle(0)
        pd.SetTime(0.0)
        pd.Save()

    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock.send_solution(mesh,  u)

    adv = FE_Evolution(m.SpMat(), k.SpMat(), b)

    ode_solver.Init(adv)
    t = 0.0
    ti = 0
    while True:
        if t > t_final - dt/2:
            break
        t, dt = ode_solver.Step(u, t, dt)
        ti = ti + 1

        if ti % vis_steps == 0:
            print("time step: " + str(ti) + ", time: " + str(np.round(t, 3)))
            if paraview:
                pd.SetCycle(ti)
                pd.SetTime(t)
                pd.Save()
            if visualization:
                sol_sock.send_solution(mesh,  u)

    u.Save('ex9-final.gf', 8)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex9')
    parser.add_argument('-m', '--mesh',
                        default='periodic-hexagon.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument("-r", "--refine",
                        default=2,
                        action='store', type=int,
                        help="Number of times to refine the mesh uniformly.")
    parser.add_argument("-p", "--problem",
                        default=0, type=int,
                        help="Problem setup to use. See options in velocity_function().")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=3, type=int,
                        help="Order (degree) of the finite elements.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument("-s", "--ode-solver",
                        default=4, type=int,
                        help="\n".join(["ODE solver: 1 - Forward Euler,\n\t",
                                        "            2 - RK2 SSP, 3 - RK3 SSP, 4 - RK4, 6 - RK6,\n\t",
                                        "            11 - Backward Euler,\n\t",
                                        "            12 - SDIRK23 (L-stable), 13 - SDIRK33,\n\t",
                                        "            22 - Implicit Midpoint Method,\n\t",
                                        "            23 - SDIRK23 (A-stable), 24 - SDIRK34", ]))
    parser.add_argument("-tf", "--t-final",
                        default=10, type=float,
                        help="Final time; start time is 0.")
    parser.add_argument("-dt", "--time-step",
                        default=0.01, type=float,
                        help="Time step.")
    parser.add_argument("-vs", "--visualization-steps",
                        default=5, type=int,
                        help="Visualize every n-th timestep.")
    parser.add_argument("-paraview", "--paraview-datafiles",
                        default=False,
                        action='store_true',
                        help="Save data files for ParaView (paraview.org) visualization.")

    args = parser.parse_args()
    parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly
    ode_solver_type = args.ode_solver
    t_final = args.t_final
    dt = args.time_step
    vis_step = args.visualization_steps
    ref_levels = args.refine
    problem = args.problem
    paraview = args.paraview_datafiles

    run(ref_levels=ref_levels,
        order=order,
        ode_solver_type=ode_solver_type,
        t_final=t_final,
        dt=dt,
        problem=problem,
        vis_steps=vis_step,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa,
        paraview=paraview)



'''
   bessel functions  (real order, complex argument)
      iv
      kv
      jv
      hankel1
      hankel2
      yv
'''
import numpy as np
from numpy import exp, log, pi, sin, cos, sqrt, abs, cosh, sinh, sum
from math import gamma
from numba import njit, float64, int64, complex128

# machine precision for double computation
E = np.finfo(np.complex128).resolution  # 1e-15
D = -np.log10(np.finfo(np.complex128).resolution)  # 15

'''
I(n, x)
  n: real
  x: compplex
'''


def iv_smallz(n, z):
    '''
    |z| < 2*sqrt(n+1)
    '''
    a = 1
    sum1 = 0
    for k in range(20):
        sum1 = sum1 + a
        fac = 1/(k+1)/(k+n+1)*(z**2/4)
        a = a*fac
        if abs(fac) < E:
            break

    return sum1*(z/2)**n/gamma(1+n)


def iv_largez(n, z):
    '''
    good for |z| > max(1.2*D+2.4, n**2/2)
    '''
    a = -(4*n**2 - 1)/(8*z)
    b = (4*n**2 - 1)/(8*z)

    sum1 = 1
    sum2 = 1
    for i in range(15):
        k = i+1
        sum1 = sum1 + a
        sum2 = sum2 + b
        k = k+1
        a = -a*(4*n**2 - (2*k-1)**2)/k/(8*z)
        b = b*(4*n**2 - (2*k-1)**2)/k/(8*z)

    ans = exp(z)/sqrt(2*pi*z) * sum1
    if np.imag(z) >= 0:
        ans = ans + sum2*exp(1j*(n+0.5)*pi)*exp(-z)/sqrt(2*pi*z)
    else:
        ans = ans + sum2*exp(-1j*(n+0.5)*pi)*exp(-z)/sqrt(2*pi*z)

    return ans


def iv_middlez(n, z):
    nu = n - np.floor(n)
    size = 40
    data = np.zeros(size, dtype=np.complex128)
    data[-2] = 1

    for i in range(len(data)-2):
        nn = len(data)-2-i+nu
        data[-3-i] = 2*nn/z*data[-2-i] + data[-1-i]

    norm = data[0]
    for k in range(len(data)):
        if k == 0:
            norm = data[0]
            fac = 1.
        else:
            fac = fac*k
            lam = 2*(nu+k)*gamma(k+2*nu)/gamma(1+2*nu)/fac
            norm += data[k]*lam

    data /= norm
    data *= (z/2.)**nu/gamma(1+nu)*exp(z)

    return data[int(np.floor(n))]


def kv_smallz(n, z):
    '''
    |z| > 2, np.real(z) > 0
    '''
    tmp = n-np.floor(n)
    nu = tmp if tmp < 0.5 else tmp-1.

    if nu == -0.5:
        k1 = exp(-z)*sqrt(pi/2/z)
        k2 = exp(-z)*sqrt(pi/2/z)

    if nu != 0:
        g1 = (1./gamma(1-nu) - 1./gamma(1+nu))/(2*nu)
    else:
        nu1 = 1e-5
        g1 = (1/gamma(1-nu1) - 1/gamma(1+nu1))/(2.*nu1)

    g2 = (1/gamma(1-nu) + 1/gamma(1+nu))/2.0

    aa = gamma(1+nu)*gamma(1-nu)

    if abs(nu) == 0.0:
        term1 = 1.0
        term2 = 1.0
    else:
        mu = nu*log(2/z)
        term1 = cosh(mu)
        term2 = sinh(mu)/mu

    fk = aa*(g1*term1 + g2*log(2./z)*term2)
    pk = 0.5*(z/2.)**(-nu)*gamma(1+nu)
    qk = 0.5*(z/2.)**(nu)*gamma(1-nu)
    ck = 1.

    k1 = ck*fk
    k2 = ck*pk
    for i in range(30):
        k = i+1

        fk = (k*fk + pk + qk)/(k**2-nu**2)

        pk = pk/(k-nu)
        qk = qk/(k+nu)
        ck = ck*(z**2/4)/k

        k1 = k1 + ck*fk
        k2 = k2 + ck*(pk - k*fk)

    k2 = 2/z*k2

    num_rec = int(n - nu)

    if num_rec == 0:
        return k1
    if num_rec == 1:
        return k2

    for i in range(num_rec-1):
        nn = i+1+nu
        tmp = k2
        k2 = 2*nn/z*k2+k1
        k1 = tmp

    return k2


def kv_largez(n, z):
    '''
    |z| > 2, np.real(z) > 0
    '''
    tmp = n-np.floor(n)
    nu = tmp if tmp < 0.5 else tmp-1.

    if nu == -0.5:
        k1 = exp(-z)*sqrt(pi/2/z)
        k2 = k1
    else:
        size = 50
        data = np.zeros(size, np.complex128)
        data[1] = 1.

        for i in range(size-2):
            nn = size-i-2.

            an = ((nn-0.5)**2 - nu**2)/nn/(nn+1)
            bn = 2*(nn+z)/(nn+1)

            data[i+2] = -(data[i] - bn*data[i+1])/an

        data /= sum(data)
        data *= (2*z)**(-nu-0.5)

        fac = cos(pi*nu)/pi*gamma(nu+0.5)*gamma(0.5-nu)
        uu0 = data[-1]/fac

        k1 = sqrt(pi)*(2*z)**nu * exp(-z)*uu0
        k2 = k1*(nu+0.5+z - data[-2]/data[-1])/z

    num_rec = int(n - nu)

    if num_rec == 0:
        return k1
    if num_rec == 1:
        return k2

    for i in range(num_rec-1):
        nn = i+1+nu
        tmp = k2
        k2 = 2*nn/z*k2+k1
        k1 = tmp

    return k2


def kv_righthalf(n, z):
    if abs(z) < 2:
        kval = kv_smallz(abs(n), z)
    else:
        kval = kv_largez(abs(n), z)
    return kval


def iv_righthalf(n, z):
    safe_guard = 1.0  # must be >1, expand intermidate region
    if abs(z) < 2*sqrt(abs(n)+1)/safe_guard:
        ival = iv_smallz(abs(n), z)
    elif abs(z) > (1.2*D+2.4)*safe_guard and abs(z) > (abs(n)**2/2)*safe_guard:
        ival = iv_largez(abs(n), z)
    else:
        ival = iv_middlez(abs(n), z)

    if n < 0:
        kval = kv_righthalf(n, z)
        ival = ival + 2./pi*sin(-n*pi)*kval

    if np.imag(z) == 0:
        return np.real(ival)

    return ival


def jv_righthalf(n, z):
    jval = exp(1j*n*pi/2.)*iv_righthalf(n, z/1j)
    if np.imag(z) == 0:
        return np.real(jval)
    return jval

#
#   kv, iv, jv, hankel1, hankel2, yv
#


def kv(n, z):
    if np.real(z) < 0:
        zz = -z
        ival = iv_righthalf(n, zz)
    else:
        zz = z
        ival = 0j

    kval = kv_righthalf(n, zz)

    ee = exp(1j*pi*n)
    if (n-np.floor(n)) == 0.0:
        ee = np.real(ee)

    if np.real(z) < 0 and np.imag(z) >= 0:
        return 1/ee*kval - 1j*pi*ival
    elif np.real(z) < 0 and np.imag(z) < 0:
        return ee*kval + 1j*pi*ival
    else:
        return kval


def iv(n, z):
    if np.real(z) < 0:
        zz = -z
    else:
        zz = z
    ival = iv_righthalf(n, zz)

    ee = exp(1j*pi*n)
    if (n-np.floor(n)) == 0.0:
        ee = np.real(ee)

    if -np.real(z) > 0 and np.imag(z) >= 0:
        return ee*ival
    elif -np.real(z) > 0 and np.imag(z) < 0:
        return 1/ee*ival

    return ival


def jv(n, z):
    if -np.real(z) > 0:
        zz = -z
    else:
        zz = z
    jval = jv_righthalf(n, zz)

    ee = exp(1j*pi*n)
    if (n-np.floor(n)) == 0.0:
        ee = np.real(ee)

    if -np.real(z) > 0 and np.imag(z) >= 0:
        return ee*jval
    elif -np.real(z) > 0 and np.imag(z) < 0:
        return 1/ee*jval

    return jval


def hankel1(n, z):
    zz = z*exp(-1j*pi/2)
    return 2/pi/1j*exp(-1j*pi*n/2)*kv(n, zz)


def hankel2(n, z):
    zz = z*exp(1j*pi/2)
    return -2/pi/1j*exp(1j*pi*n/2)*kv(n, zz)


def yv(n, z):
    return (hankel1(n, z) - hankel2(n, z))/2/1j


jitter = njit(complex128(float64, complex128))

iv_smallz = jitter(iv_smallz)
iv_largez = jitter(iv_largez)
iv_middlez = jitter(iv_middlez)
kv_smallz = jitter(kv_smallz)
kv_largez = jitter(kv_largez)
kv_righthalf = jitter(kv_righthalf)
iv_righthalf = jitter(iv_righthalf)
jv_righthalf = jitter(jv_righthalf)
kv = jitter(kv)
iv = jitter(iv)
jv = jitter(jv)
hankel1 = jitter(hankel1)
hankel2 = jitter(hankel2)
yv = jitter(yv)



'''
   MFEM example 3 parallel version

   See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.par as mfem
from mfem.par import intArray
from mpi4py import MPI
import numpy as np
from numpy import sin, array


static_cond = False
order = 1

nprc = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '{:0>6d}'.format(myid)
verbose = (myid == 0)


def run(order=1,
        static_cond=False,
        freq=1,
        meshfile='',
        visualization=False,
        device='cpu',
        numba=False,
        pa=False):

    kappa = np.pi*freq

    # 3. Enable hardware devices such as GPUs, and programming models such as
    #    CUDA, OCCA, RAJA and OpenMP based on command line options.
    device = mfem.Device(device)
    if myid == 0:
        device.Print()

    # 4. Read the (serial) mesh from the given mesh file on all processors.  We
    #    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
    #    and volume meshes with the same code.
    mesh = mfem.Mesh(meshfile, 1, 1)

    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    if numba:
        @mfem.jit.vector(sdim=sdim, shape=(dim,))
        def E_exact(x):
            if shape[0] == 3:
                return np.array((sin(kappa*x[1]),
                                 sin(kappa*x[2]),
                                 sin(kappa*x[0])))
            else:
                if sdim == 3:
                    return np.array([sin(kappa*x[1]),
                                     sin(kappa*x[0]),
                                     0.0])
                else:
                    return np.array([sin(kappa*x[1]),
                                     sin(kappa*x[0])])

        @mfem.jit.vector(sdim=sdim, shape=(dim,), interface="c++")
        def f_exact(x, out):
            if shape[0] == 3:
                out[0] = (1 + kappa**2)*sin(kappa * x[1])
                out[1] = (1 + kappa**2)*sin(kappa * x[2])
                out[2] = (1 + kappa**2)*sin(kappa * x[0])
            else:
                out[0] = (1 + kappa**2)*sin(kappa * x[1])
                out[1] = (1 + kappa**2)*sin(kappa * x[0])
                if shape[0] == 3:
                    out[2] = 0.

    else:
        class cE_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                if dim == 3:
                    return (sin(kappa * x[1]),
                            sin(kappa * x[2]),
                            sin(kappa * x[0]))
                elif sdim == 2:
                    return (sin(kappa * x[1]),
                            sin(kappa * x[0]),)
                else:
                    return (sin(kappa * x[1]),
                            sin(kappa * x[0]),
                            0)

        E_exact = cE_exact()

        class cf_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                if dim == 3:
                    return ((1 + kappa**2)*sin(kappa * x[1]),
                            (1 + kappa**2)*sin(kappa * x[2]),
                            (1 + kappa**2)*sin(kappa * x[0]))
                elif sdim == 2:
                    return ((1 + kappa**2)*sin(kappa * x[1]),
                            (1 + kappa**2)*sin(kappa * x[0]),)
                else:
                    return ((1 + kappa**2)*sin(kappa * x[1]),
                            (1 + kappa**2)*sin(kappa * x[0]),
                            0.0)

        f_exact = cf_exact()

    # 5. Refine the serial mesh on all processors to increase the resolution. In
    #    this example we do 'ref_levels' of uniform refinement. We choose
    #    'ref_levels' to be the largest number that gives a final mesh with no
    #    more than 1,000 elements

    ref_levels = int(np.floor(np.log(1000./mesh.GetNE())/np.log(2.)/dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()

    # 6. Define a parallel mesh by a partitioning of the serial mesh. Refine
    #    this mesh further in parallel to increase the resolution. Once the
    #    parallel mesh is defined, the serial mesh can be deleted.
    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    par_ref_levels = 2
    for l in range(par_ref_levels):
        pmesh.UniformRefinement()

    # 7. Define a parallel finite element space on the parallel mesh. Here we
    #    use the Nedelec finite elements of the specified order.
    fec = mfem.ND_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    size = fespace.GlobalTrueVSize()

    if verbose:  # note that size should be evaulated on all nodes
        print("Number of finite element unknowns: " + str(size))

    # 8. Determine the list of true (i.e. parallel conforming) essential
    #    boundary dofs. In this example, the boundary conditions are defined
    #    by marking all the boundary attributes from the mesh as essential
    #    (Dirichlet) and converting them to a list of true dofs.
    ess_tdof_list = intArray()
    if mesh.bdr_attributes.Size():
        ess_bdr = intArray(mesh.bdr_attributes.Max())
        ess_bdr.Assign(1)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 9. Set up the parallel linear form b(.) which corresponds to the
    #    right-hand side of the FEM linear system, which in this case is
    #    (f,phi_i) where f is given by the function f_exact and phi_i are the
    #    basis functions in the finite element fespace.
    b = mfem.ParLinearForm(fespace)
    dd = mfem.VectorFEDomainLFIntegrator(f_exact)
    b.AddDomainIntegrator(dd)
    b.Assemble()

    # 10. Define the solution vector x as a parallel finite element grid function
    #     corresponding to fespace. Initialize x by projecting the exact
    #     solution. Note that only values from the boundary edges will be used
    #     when eliminating the non-homogeneous boundary condition to modify the
    #     r.h.s. vector b.
    x = mfem.ParGridFunction(fespace)
    x.ProjectCoefficient(E_exact)

    # 11. Set up the parallel bilinear form corresponding to the EM diffusion
    #     operator curl muinv curl + sigma I, by adding the curl-curl and the
    #     mass domain integrators.
    muinv = mfem.ConstantCoefficient(1.0)
    sigma = mfem.ConstantCoefficient(1.0)
    a = mfem.ParBilinearForm(fespace)
    if pa:
        a.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(muinv))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(sigma))

    # 12. Assemble the parallel bilinear form and the corresponding linear
    #     system, applying any necessary transformations such as: parallel
    #     assembly, eliminating boundary conditions, applying conforming
    #     constraints for non-conforming AMR, static condensation, etc
    if (static_cond):
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

    # 13. Solve the system AX=B using PCG with an AMS preconditioner.
    if pa:
        ams = mfem.MatrixFreeAMS(a, A, fespace, muinv, sigma, None, ess_bdr)
        cg = mfem.CGSolver(MPI.COMM_WORLD)
        cg.SetRelTol(1e-12)
        cg.SetMaxIter(1000)
        cg.SetPrintLevel(1)
        cg.SetOperator(A.Ptr())
        cg.SetPreconditioner(ams)
        cg.Mult(B, X)
    else:
        prec_fespace = (a.SCParFESpace() if a.StaticCondensationIsEnabled()
                        else fespace)
        Amat = A.AsHypreParMatrix()

        if verbose:
            print("Size of linear system: " + str(Amat.GetGlobalNumRows()))

        ams = mfem.HypreAMS(Amat, prec_fespace)
        pcg = mfem.HyprePCG(Amat)
        pcg.SetTol(1e-12)
        pcg.SetMaxIter(500)
        pcg.SetPrintLevel(2)
        pcg.SetPreconditioner(ams)
        pcg.Mult(B, X)

    # 14. Recover the parallel grid function corresponding to X. This is the
    #     local finite element solution on each processor.
    a.RecoverFEMSolution(X, b, x)

    # 15. Compute and print the L^2 norm of the error.
    err = x.ComputeL2Error(E_exact)
    if verbose:  # note that err should be evaulated on all nodes
        print("|| E_h - E ||_{L^2} = " + "{:g}".format(err))

    # 16. Save the refined mesh and the solution in parallel. This output can
    #     be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".
    x.Save('sol.'+smyid, 8)
    pmesh.Print('mesh.'+smyid)

    # 17. Send the solution by socket to a GLVis server
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)
        sol_sock << "parallel " << nprc << " " << myid << "\n"
        sol_sock << "solution\n" << pmesh << x
        sol_sock << "window_title 'Soluiton real part'"
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex3 (Definite Maxwell Problem)')
    parser.add_argument('-m', '--mesh',
                        default="beam-tet.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=1.0,
                        help="Set the frequency for the exact")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        HAS_NUMBA = False
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)

    if verbose:
        parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation

    from os.path import expanduser, join, dirname
    meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))

    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly
    freq = args.frequency
    numba = args.numba

    run(freq=freq,
        order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa,
        numba=numba)



import numpy as np
from scipy.sparse import csr_matrix

import mfem.ser as mfem


smat = csr_matrix([[1,2,3,], [0, 0, 1], [2, 0, 0]])
print(mfem.SparseMatrix(smat))
print(smat)
mmat = mfem.SparseMatrix(smat)
mmat.Print()



'''
    MFEM Example 18 - Serial/Parallel Shared Code

    This is a python translation of ex18.hpp


'''
import numpy as np
from mfem import mfem_mode

if mfem_mode is None or mfem_mode == 'serial':
    import mfem.ser as mfem
else:
    import mfem.par as mfem

from os.path import expanduser, join, dirname


class DGHyperbolicConservationLaws(mfem.TimeDependentOperator):
    def __init__(self, vfes_, formIntegrator_, preassembleWeakDivergence=True):

        super(DGHyperbolicConservationLaws, self).__init__(
            vfes_.GetTrueVSize())
        self.num_equations = formIntegrator_.num_equations
        self.vfes = vfes_
        self.dim = vfes_.GetMesh().SpaceDimension()
        self.formIntegrator = formIntegrator_

        self.z = mfem.Vector(vfes_.GetTrueVSize())

        self.weakdiv = None
        self.max_char_speed = None

        self.ComputeInvMass()

        if mfem_mode == 'serial':
            self.nonlinearForm = mfem.NonlinearForm(self.vfes)
        else:
            if isinstance(self.vfes, mfem.ParFiniteElementSpace):
                self.nonlinearForm = mfem.ParNonlinearForm(self.vfes)
            else:
                self.nonlinearForm = mfem.NonlinearForm(self.vfes)

        if preassembleWeakDivergence:
            self.ComputeWeakDivergence()
        else:
            self.nonlinearForm.AddDomainIntegrator(self.formIntegrator)

        self.nonlinearForm.AddInteriorFaceIntegrator(self.formIntegrator)
        self.nonlinearForm.UseExternalIntegrators()

    def GetMaxCharSpeed(self):
        return self.max_char_speed

    def ComputeInvMass(self):
        inv_mass = mfem.InverseIntegrator(mfem.MassIntegrator())

        self.invmass = [None]*self.vfes.GetNE()
        for i in range(self.vfes.GetNE()):
            dof = self.vfes.GetFE(i).GetDof()
            self.invmass[i] = mfem.DenseMatrix(dof)
            inv_mass.AssembleElementMatrix(self.vfes.GetFE(i),
                                           self.vfes.GetElementTransformation(
                                               i),
                                           self.invmass[i])

    def ComputeWeakDivergence(self):
        weak_div = mfem.TransposeIntegrator(mfem.GradientIntegrator())

        weakdiv_bynodes = mfem.DenseMatrix()

        self.weakdiv = [None]*self.vfes.GetNE()

        for i in range(self.vfes.GetNE()):
            dof = self.vfes.GetFE(i).GetDof()
            weakdiv_bynodes.SetSize(dof, dof*self.dim)
            weak_div.AssembleElementMatrix2(self.vfes.GetFE(i),
                                            self.vfes.GetFE(i),
                                            self.vfes.GetElementTransformation(
                                                i),
                                            weakdiv_bynodes)
            self.weakdiv[i] = mfem.DenseMatrix()
            self.weakdiv[i].SetSize(dof, dof*self.dim)

            # Reorder so that trial space is ByDim.
            # This makes applying weak divergence to flux value simpler.
            for j in range(dof):
                for d in range(self.dim):
                    self.weakdiv[i].SetCol(
                        j*self.dim + d, weakdiv_bynodes.GetColumn(d*dof + j))

    def Mult(self, x, y):
        # 0. Reset wavespeed computation before operator application.
        self.formIntegrator.ResetMaxCharSpeed()

        # 1. Apply Nonlinear form to obtain an auxiliary result
        #         z = - <F̂(u_h,n), [[v]]>_e
        #    If weak-divergence is not preassembled, we also have weak-divergence
        #         z = - <F̂(u_h,n), [[v]]>_e + (F(u_h), ∇v)
        self.nonlinearForm.Mult(x, self.z)
        #print("!!!!", self.weakdiv)
        if self.weakdiv is not None:  # if weak divergence is pre-assembled
            # Apply weak divergence to F(u_h), and inverse mass to z_loc + weakdiv_loc

            current_state = mfem.Vector()     # view of current state at a node
            current_flux = mfem.DenseMatrix()  # flux of current state

            # element flux value. Whose column is ordered by dim.
            flux = mfem.DenseMatrix()
            # view of current states in an element, dof x num_eq
            current_xmat = mfem.DenseMatrix()
            # view of element auxiliary result, dof x num_eq
            current_zmat = mfem.DenseMatrix()
            current_ymat = mfem.DenseMatrix()  # view of element result, dof x num_eq

            fluxFunction = self.formIntegrator.GetFluxFunction()

            xval = mfem.Vector()
            zval = mfem.Vector()
            flux_vec = mfem.Vector()

            for i in range(self.vfes.GetNE()):
                Tr = self.vfes.GetElementTransformation(i)
                dof = self.vfes.GetFE(i).GetDof()
                vdofs = mfem.intArray(self.vfes.GetElementVDofs(i))

                x.GetSubVector(vdofs, xval)
                current_xmat.UseExternalData(
                    xval.GetData(), dof, self.num_equations)

                #
                # Python Note:
                #    C++ code access to array data with offset is done bu GetData() + offset
                #    In Python, the same can be done by using numpy array generated from Vector::GetDataArray(),
                #
                #       array = vec.GetDataArray()
                #       new_data_pointer = mfem.Vector(array[10:]).GetData()
                #
                #    note that the above does not work if mfem.Vector is replaced by mfem.DenseMatrix
                #    This is because, while MFEM stores data in colume-major, Python numpy store raw-major.
                #

                flux.SetSize(self.num_equations, self.dim*dof)
                flux_vec = mfem.Vector(
                    flux.GetData(), self.num_equations*self.dim*dof)
                data = flux_vec.GetDataArray()

                for j in range(dof):  # compute flux for all nodes in the element
                    current_xmat.GetRow(j, current_state)

                    data_ptr = mfem.Vector(
                        data[self.num_equations*self.dim*j:]).GetData()
                    current_flux = mfem.DenseMatrix(data_ptr,
                                                    self.num_equations, dof)
                    fluxFunction.ComputeFlux(current_state, Tr, current_flux)

                # Compute weak-divergence and add it to auxiliary result, z
                # Recalling that weakdiv is reordered by dim, we can apply
                # weak-divergence to the transpose of flux.
                self.z.GetSubVector(vdofs, zval)
                current_zmat.UseExternalData(
                    zval.GetData(), dof, self.num_equations)
                mfem.AddMult_a_ABt(1.0, self.weakdiv[i], flux, current_zmat)

                # Apply inverse mass to auxiliary result to obtain the final result
                current_ymat.SetSize(dof, self.num_equations)
                mfem.Mult(self.invmass[i], current_zmat, current_ymat)
                y.SetSubVector(vdofs, current_ymat.GetData())

        else:

            # Apply block inverse mass
            zval = mfem.Vector()  # / z_loc, dof*num_eq

            # view of element auxiliary result, dof x num_eq
            current_zmat = mfem.DenseMatrix()
            current_ymat = mfem.DenseMatrix()  # view of element result, dof x num_eq

            for i in range(self.vfes.GetNE()):
                dof = self.vfes.GetFE(i).GetDof()
                vdofs = mfem.intArray(self.vfes.GetElementVDofs(i))
                self.z.GetSubVector(vdofs, zval)
                current_zmat.UseExternalData(
                    zval.GetData(), dof, self.num_equations)
                current_ymat.SetSize(dof, self.num_equations)
                mfem.Mult(self.invmass[i], current_zmat, current_ymat)
                y.SetSubVector(vdofs, current_ymat.GetData())

        self.max_char_speed = self.formIntegrator.GetMaxCharSpeed()

    def Update(self):
        self.nonlinearForm.Update()
        height = self.nonlinearForm.Height()
        width = height
        self.z.SetSize(height)

        ComputeInvMass()
        if self.weakdiv is None:
            self.ComputeWeakDivergence()


def GetMovingVortexInit(radius, Minf, beta, gas_constant, specific_heat_ratio):
    def func(x, y):
        xc = 0.0
        yc = 0.0
        # Nice units
        vel_inf = 1.
        den_inf = 1.

        pres_inf = (den_inf / specific_heat_ratio) * \
            (vel_inf / Minf) * (vel_inf / Minf)
        temp_inf = pres_inf / (den_inf * gas_constant)

        r2rad = 0.0
        r2rad += (x[0] - xc) * (x[0] - xc)
        r2rad += (x[1] - yc) * (x[1] - yc)
        r2rad /= (radius * radius)

        shrinv1 = 1.0 / (specific_heat_ratio - 1.)

        velX = vel_inf * \
            (1 - beta * (x[1] - yc) / radius * np.exp(-0.5 * r2rad))
        velY = vel_inf * beta * (x[0] - xc) / radius * np.exp(-0.5 * r2rad)
        vel2 = velX * velX + velY * velY

        specific_heat = gas_constant * specific_heat_ratio * shrinv1

        temp = temp_inf - (0.5 * (vel_inf * beta) *
                           (vel_inf * beta) / specific_heat * np.exp(-r2rad))

        den = den_inf * (temp/temp_inf)**shrinv1
        pres = den * gas_constant * temp
        energy = shrinv1 * pres / den + 0.5 * vel2

        y[0] = den
        y[1] = den * velX
        y[2] = den * velY
        y[3] = den * energy

    return func


def EulerMesh(meshfile, problem):
    if meshfile == '':
        if problem in (1, 2, 3):
            meshfile = "periodic-square.mesh"

        elif problem == 4:
            meshfile = "periodic-segment.mesh"

        else:
            assert False, "Default mesh file not given for problem = " + \
                str(problem)

        meshfile = expanduser(join(dirname(__file__), '..', 'data', meshfile))

    return mfem.Mesh(meshfile, 1, 1)

# Initial condition


def EulerInitialCondition(problem, specific_heat_ratio, gas_constant):

    if problem == 1:
        # fast moving vortex
        func = GetMovingVortexInit(0.2, 0.5, 1. / 5., gas_constant,
                                   specific_heat_ratio)
        return mfem.jit.vector(vdim=4, interface="c++")(func)

    elif problem == 2:
        # slow moving vortex
        func = GetMovingVortexInit(0.2, 0.05, 1. / 50., gas_constant,
                                   specific_heat_ratio)
        return mfem.jit.vector(vdim=4, interface="c++")(func)

    elif problem == 3:
        # moving sine wave
        @ mfem.jit.vector(vdim=4, interface="c++")
        def func(x, y):
            assert len(x) > 2, "2D is not supportd for this probl"
            density = 1.0 + 0.2 * np.sin(np.pi*(x[0]+x[1]))
            velocity_x = 0.7
            velocity_y = 0.3
            pressure = 1.0
            energy = (pressure / (1.4 - 1.0) +
                      density * 0.5 * (velocity_x * velocity_x + velocity_y * velocity_y))

            y[0] = density
            y[1] = density * velocity_x
            y[2] = density * velocity_y
            y[3] = energy

        return func

    elif problem == 4:
        @ mfem.jit.vector(vdim=3, interface="c++")
        def func(x, y):
            density = 1.0 + 0.2 * np.sin(np.pi * 2 * x[0])
            velocity_x = 1.0
            pressure = 1.0
            energy = pressure / (1.4 - 1.0) + density * \
                0.5 * (velocity_x * velocity_x)

            y[0] = density
            y[1] = density * velocity_x
            y[2] = energy

        return func

    else:
        assert False, "Problem Undefined"



'''
   MFEM example 4

   See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join, dirname, exists
import numpy as np
from numpy import sin, array, cos

set_bc = True
static_cond = False
hybridization = False

freq = 1.0
kappa = np.pi*freq

order = 1


meshfile = expanduser(join(path, 'data', 'star.mesh'))
if not exists(meshfile):
    path = dirname(dirname(__file__))
    path = dirname(path)    
    meshfile = expanduser(join(path, 'data', 'star.mesh'))

mesh = mfem.Mesh(meshfile, 1,1)
dim = mesh.Dimension()
sdim = mesh.SpaceDimension()
ndim = mesh.Dimension()

from numba import cfunc, carray    

@cfunc(mfem.vector_sig)
def E_exact(p, out, sdim, vdim):
    out_array = carray(out, (vdim, ))    
    x = p[0]; y = p[1]
    F0 = cos(kappa*x)*sin(kappa*y)
    F1 = cos(kappa*y)*sin(kappa*x)
    out_array[0] = F0
    out_array[1] = F1
    if vdim == 3:
        out_array[2] = 0.0
        
@cfunc(mfem.vector_sig)        
def f_exact(p, out, sdim, vdim):
    out_array = carray(out, (vdim, ))    
    x = p[0]; y = p[1]
    temp = 1. + 2.*kappa*kappa
       
    F0 = temp * cos(kappa*x)*sin(kappa*y)
    F1 = temp * cos(kappa*y)*sin(kappa*x)
    out_array[0] = F0
    out_array[1] = F1
    if vdim == 3:
        out_array[2] = 0.0

ref_levels = int(np.floor(np.log(25000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
   mesh.UniformRefinement();

fec = mfem.RT_FECollection(order-1, dim)
fespace = mfem.FiniteElementSpace(mesh, fec)

print("Number of finite element unknows : " + str(fespace.GetTrueVSize()))

ess_tdof_list = intArray();
if mesh.bdr_attributes.Size():
    ess_bdr = intArray(mesh.bdr_attributes.Max())
    if set_bc: ess_bdr.Assign(1)
    else: ess_bdr.Assign(0)
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list);

b = mfem.LinearForm(fespace);
f = mfem.VectorNumbaFunction(f_exact, sdim, ndim).GenerateCoefficient()
dd = mfem.VectorFEDomainLFIntegrator(f)
b.AddDomainIntegrator(dd)
b.Assemble()
    
x = mfem.GridFunction(fespace)
F = mfem.VectorNumbaFunction(E_exact, sdim, ndim).GenerateCoefficient()
x.ProjectCoefficient(F);

alpha = mfem.ConstantCoefficient(1.0);
beta = mfem.ConstantCoefficient(1.0);
a = mfem.BilinearForm(fespace);
a.AddDomainIntegrator(mfem.DivDivIntegrator(alpha));
a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(beta));


if (static_cond):  a.EnableStaticCondensation()
elif (hybridization):
    hfec = mfem.DG_Interface_FECollection(order-1, dim)
    hfes = mfem.FiniteElementSpace(mesh, hfec)
    a.EnableHybridization(hfes, mfem.NormalTraceJumpIntegrator(),
                          ess_tdof_list)
a.Assemble();


A = mfem.OperatorPtr()
B = mfem.Vector()
X = mfem.Vector()
a.FormLinearSystem(ess_tdof_list, x, b, A, X, B);
## Here, original version calls hegith, which is not
## defined in the header...!?
print("Size of linear system: " + str(A.Height())) 

# 10. Solve
AA = mfem.OperatorHandle2SparseMatrix(A)
M = mfem.GSSmoother(AA)
X.Print("x")
mfem.PCG(AA, M, B, X, 1, 10000, 1e-20, 0.0);

# 11. Recover the solution as a finite element grid function.
a.RecoverFEMSolution(X, b, x);

print("|| F_h - F ||_{L^2} = " + str(x.ComputeL2Error(F)))

mesh.Print('refined.mesh', 8)
x.Save('sol.gf', 8)



'''
   MFEM example 11

   See c++ version in the MFEM library for more detail 

   How to run:
      mpirun -np 2 python <arguments>
  
   Example of arguments:
      ex11p.py -m square-disc.mesh
      ex11p.py -m star.mesh
      ex11p.py -m escher.mesh
      ex11p.py -m fichera.mesh
      ex11p.py -m square-disc-p2.vtk -o 2
      ex11p.py -m square-disc-p3.mesh -o 3
      ex11p.py -m square-disc-nurbs.mesh -o -1
      ex11p.py -m disc-nurbs.mesh -o -1 -n 20
      ex11p.py -m pipe-nurbs.mesh -o -1
      ex11p.py -m ball-nurbs.mesh -o 2
      ex11p.py -m star-surf.mesh
      ex11p.py -m square-disc-surf.mesh
      ex11p.py -m inline-segment.mesh
      ex11p.py -m amr-quad.mesh
      ex11p.py -m amr-hex.mesh
      ex11p.py -m mobius-strip.mesh -n 8
      ex11p.py -m klein-bottle.mesh -n 10
'''
import sys
from os.path import expanduser, join, dirname
import numpy as np

from mfem.common.arg_parser import ArgParser
import mfem.par as mfem
from mpi4py import MPI


num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


parser = ArgParser(description='Ex11 ')
parser.add_argument('-m', '--mesh',
                    default='star.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-rs', '--refine-serial',
                    default=2,
                    action='store', type=int,
                    help="Number of times to refine the mesh uniformly in serial.")

parser.add_argument('-rp', '--refine-parallel',
                    default=1,
                    action='store', type=int,
                    help="Number of times to refine the mesh uniformly in parallel.")
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help=("Finite element order (polynomial degree) or -1 for isoparametric space."))
parser.add_argument("-n", "--num-eigs",
                    action='store', default=5, type=int,
                    help="Number of desired eigenmodes.")
parser.add_argument("-sp", "--strumpack",
                    action='store_true', default=False,
                    help="Use the STRUMPACK Solver.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')
args = parser.parse_args()


ser_ref_levels = args.refine_serial
par_ref_levels = args.refine_parallel
order = args.order
nev = args.num_eigs
visualization = args.visualization
use_strumpack = args.strumpack
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the mesh from the given mesh file on all processors. We can handle
#    triangular, quadrilateral, tetrahedral, hexahedral, surface and volume
#    meshes with the same code
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Refine the serial mesh on all processors to increase the resolution. In
#    this example we do 'ref_levels' of uniform refinement (2 by default, or
#    specified on the command line with -rs).
for x in range(ser_ref_levels):
    mesh.UniformRefinement()

# 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution (1 time by
#    default, or specified on the command line with -rp). Once the parallel
#    mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
for l in range(par_ref_levels):
    pmesh.UniformRefinement()

# 6. Define a parallel finite element space on the parallel mesh. Here we
#    use continuous Lagrange finite elements of the specified order. If
#    order < 1, we instead use an isoparametric/isogeometric space.
if order > 0:
    fec = mfem.H1_FECollection(order, dim)
elif pmesh.GetNodes():
    fec = pmesh.GetNodes().OwnFEC()
else:
    fec = mfem.H1_FECollection(1, dim)

fespace = mfem.ParFiniteElementSpace(pmesh, fec)
fe_size = fespace.GlobalTrueVSize()

if (myid == 0):
    print('Number of unknowns: ' + str(fe_size))

# 7. Set up the parallel bilinear forms a(.,.) and m(.,.) on the finite
#    element space. The first corresponds to the Laplacian operator -Delta,
#    while the second is a simple mass matrix needed on the right hand side
#    of the generalized eigenvalue problem below. The boundary conditions
#    are implemented by elimination with special values on the diagonal to
#    shift the Dirichlet eigenvalues out of the computational range. After
#    serial and parallel assembly we extract the corresponding parallel
#    matrices A and M.
one = mfem.ConstantCoefficient(1.0)

ess_bdr = mfem.intArray()
if pmesh.bdr_attributes.Size() != 0:
    ess_bdr.SetSize(pmesh.bdr_attributes.Max())
    ess_bdr.Assign(1)

a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
if pmesh.bdr_attributes.Size() == 0:
    # Add a mass term if the mesh has no boundary, e.g. periodic mesh or
    # closed surface.
    a.AddDomainIntegrator(mfem.MassIntegrator(one))

a.Assemble()
a.EliminateEssentialBCDiag(ess_bdr, 1.0)
a.Finalize()


m = mfem.ParBilinearForm(fespace)
m.AddDomainIntegrator(mfem.MassIntegrator(one))
m.Assemble()

# shift the eigenvalue corresponding to eliminated dofs to a large value
m.EliminateEssentialBCDiag(ess_bdr,  3.0e-300)
m.Finalize()

A = a.ParallelAssemble()
M = m.ParallelAssemble()

if use_strumpack:
    import mfem.par.strumpack as strmpk
    Arow = strmpk.STRUMPACKRowLocMatrix(A)


# 8. Define and configure the LOBPCG eigensolver and the BoomerAMG
#    preconditioner for A to be used within the solver. Set the matrices
#    which define the generalized eigenproblem A x = lambda M x.
#    We don't support SuperLU

if use_strumpack:
    args = ["--sp_hss_min_sep_size", "128", "--sp_enable_hss"]
    strumpack = strmpk.STRUMPACKSolver(args, MPI.COMM_WORLD)
    strumpack.SetPrintFactorStatistics(True)
    strumpack.SetPrintSolveStatistics(False)
    strumpack.SetKrylovSolver(strmpk.KrylovSolver_DIRECT)
    strumpack.SetReorderingStrategy(strmpk.ReorderingStrategy_METIS)
    strumpack.SetMC64Job(strmpk.MC64Job_NONE)
    # strumpack.SetSymmetricPattern(True)
    strumpack.SetOperator(Arow)
    strumpack.SetFromCommandLine()
    precond = strumpack
else:
    amg = mfem.HypreBoomerAMG(A)
    amg.SetPrintLevel(0)
    precond = amg

lobpcg = mfem.HypreLOBPCG(MPI.COMM_WORLD)
lobpcg.SetNumModes(nev)
lobpcg.SetPreconditioner(precond)
lobpcg.SetMaxIter(200)
lobpcg.SetTol(1e-8)
lobpcg.SetPrecondUsageMode(1)
lobpcg.SetPrintLevel(1)
lobpcg.SetMassMatrix(M)
lobpcg.SetOperator(A)

# 9. Compute the eigenmodes and extract the array of eigenvalues. Define a
#    parallel grid function to represent each of the eigenmodes returned by
#    the solver.
eigenvalues = mfem.doubleArray()
lobpcg.Solve()
lobpcg.GetEigenvalues(eigenvalues)
x = mfem.ParGridFunction(fespace)

# 10. Save the refined mesh and the modes in parallel. This output can be
#     viewed later using GLVis: "glvis -np <np> -m mesh -g mode".

smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
pmesh.Print(mesh_name, 8)

for i in range(nev):
    x.Assign(lobpcg.GetEigenvector(i))
    sol_name = "mode_"+str(i).zfill(2)+"."+smyid
    x.Save(sol_name, 8)

# 11. Send the solution by socket to a GLVis server.
if (visualization):
    mode_sock = mfem.socketstream("localhost", 19916)
    mode_sock.precision(8)

    for i in range(nev):
        if (myid == 0):
            print("Eigenmode " + str(i+1) + '/' + str(nev) +
                  ", Lambda = " + str(eigenvalues[i]))

        # convert eigenvector from HypreParVector to ParGridFunction
        x.Assign(lobpcg.GetEigenvector(i))

        mode_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
        mode_sock.send_solution(pmesh,   x)
        mode_sock.send_text("window_title 'Eigenmode " + str(i+1) + '/' +
                            str(nev) + ", Lambda = " + str(eigenvalues[i]) + "'")

        c = None
        if (myid == 0):
            from builtins import input
            c = input("press (q)uit or (c)ontinue --> ")
        c = MPI.COMM_WORLD.bcast(c, root=0)
        if (c != 'c'):
            break

    mode_sock.close()



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def run_test():
    print("Test complex_operator module")
    Nvert = 6; Nelem = 8; Nbelem=2
    
    mesh = mfem.Mesh(2, Nvert, Nelem, 2, 3)
    tri_v = [[1.,  0.,  0.], [0.,  1.,  0.], [-1.,  0.,  0.],
             [0., -1.,  0.], [0.,  0.,  1.], [ 0.,  0., -1.]]
    tri_e = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4],
             [1, 0, 5], [2, 1, 5], [3, 2, 5], [0, 3, 5]]
    tri_l = [[1,4], [1,2]]
    
    for j in range(Nvert):
        mesh.AddVertex(tri_v[j])
    for j in range(Nelem):
        mesh.AddTriangle(tri_e[j], 1)
    for j in range(Nbelem):
        mesh.AddBdrSegment(tri_l[j], 1)
        
    mesh.FinalizeTriMesh(1,1, True)
    dim = mesh.Dimension()
    order = 1
    fec = mfem.H1_FECollection(order, dim)

    if use_parallel:
        mesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)        
        fes = mfem.ParFiniteElementSpace(mesh, fec)        
        a1 = mfem.ParBilinearForm(fes)
        a2 = mfem.ParBilinearForm(fes)    
    else:
        fes = mfem.FiniteElementSpace(mesh, fec)        
        a1 = mfem.BilinearForm(fes)
        a2 = mfem.BilinearForm(fes)    
    one = mfem.ConstantCoefficient(1.0)    
    a1.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
    a1.Assemble()
    a1.Finalize();

    a2.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
    a2.Assemble()
    a2.Finalize()
    
    if use_parallel:
        M1 = a1.ParallelAssemble()
        M2 = a2.ParallelAssemble()
        M1.Print('M1')
        width = fes.GetTrueVSize()
        #X = mfem.HypreParVector(fes)
        #Y = mfem.HypreParVector(fes)
        #X.SetSize(fes.TrueVSize())
        #Y.SetSize(fes.TrueVSize())
        #from mfem.common.parcsr_extra import ToScipyCoo
        #MM1 = ToScipyCoo(M1)
        #print(MM1.toarray())
        #print(MM1.dot(np.ones(6)))        
    else:
        M1 = a1.SpMat()        
        M2 = a2.SpMat()
        M1.Print('M1')
        width = fes.GetVSize()
        
        #X = mfem.Vector()
        #Y = mfem.Vector()
        #X.SetSize(M1.Width())
        #Y.SetSize(M1.Height())
        #from mfem.common.sparse_utils import sparsemat_to_scipycsr
        #MM1 = sparsemat_to_scipycsr(M1, np.float)
        #print(MM1.toarray())        
        #print(MM1.dot(np.ones(6)))
    #X.Assign(0.0)
    #X[0] = 1.0
    #M1.Mult(X, Y)
    #print(Y.GetDataArray())
    
    Mc = mfem.ComplexOperator(M1, M2, hermitan=True)
    offsets = mfem.intArray([0, width, width])
    offsets.PartialSum()

    x = mfem.BlockVector(offsets)
    y = mfem.BlockVector(offsets)

    x.GetBlock(0).Assign(0)
    if myid == 0:
        x.GetBlock(0)[0] = 1.0
    x.GetBlock(1).Assign(0)
    if myid == 0:
        x.GetBlock(1)[0] = 1.0
    
    Mc.Mult(x, y)
    print("x", x.GetDataArray())    
    print("y", y.GetDataArray())
    
    if myid == 0:
        x.GetBlock(1)[0] = -1.0

    x.Print()    
    Mc.Mult(x, y)
    print("x", x.GetDataArray())    
    print("y", y.GetDataArray())
    
if __name__=='__main__':
    run_test()



'''
   MFEM example 15pe

   How to run:
      mpirun -np 2 python <arguments>

   Example of arguments:
      ex15p.py
      ex15p.py -o 1 -y 0.4
      ex15p.py -o 4 -y 0.1
      ex15p.py -n 5
      ex15p.py -p 1 -n 3

      Other meshes:

      ex15p.py -m square-disc-nurbs.mesh
      ex15p.py -m disc-nurbs.mesh
      ex15p.py -m fichera.mesh -tf 0.5
      ex15p.py -m ball-nurbs.mesh -tf 0.5
      ex15p.py -m mobius-strip.mesh
      ex15p.py -m amr-quad.mesh

      Conforming meshes (no derefinement):

      ex15p.py -m square-disc.mesh
      ex15p.py -m escher.mesh -r 2 -tf 0.3

'''

import sys
from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np
from numpy import cos, sin, pi, exp, sqrt, arctan

import mfem.par as mfem
from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank

parser = ArgParser(description='Ex15p')
parser.add_argument('-m', '--mesh',
                    default='star-hilbert.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument("-p", "--problem",
                    action='store', default=0, type=int,
                    help="Problem setup to use: 0 = spherical front, 1 = ball.")
parser.add_argument("-n", "--nfeatures",
                    action='store', default=1, type=int,
                    help="Number of solution features (fronts/balls).")
parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
parser.add_argument("-e", "--max-err",
                    action='store', default=1e-4, type=float,
                    help="Maximum element error")
parser.add_argument("-y", "--hysteresis",
                    action='store', default=0.25, type=float,
                    help="Derefinement safety coefficient.")
parser.add_argument('-r', '--ref-levels',
                    action='store', default=0, type=int,
                    help="Number of inital uniform refinement")
parser.add_argument("-l", "--nc-limit",
                    action='store', default=3, type=int,
                    help="Maximum level of hanging nodes.")
parser.add_argument('-tf', '--t-final',
                    action='store', default=1.0, type=float,
                    help="Final time; start time is 0.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')

args = parser.parse_args()
ref_levels = args.ref_levels
problem = args.problem
nfeatures = args.nfeatures
order = args.order
max_elem_error = args.max_err
hysteresis = args.hysteresis
nc_limit = args.nc_limit
t_final = args.t_final
visualization = args.visualization
if myid == 0:
    parser.print_options(args)

alpha = 0.02

device = mfem.Device('cpu')
if myid == 0:
    device.Print()


def front(x, y, z, t, dim):
    r = sqrt(x**2 + y**2 + z**2)
    return exp(-0.5 * ((r - t)/alpha) ** 2)


def ball(x, y, z, t, dim):
    r = sqrt(x**2 + y**2 + z**2)
    return -atan(2.*(r - t)/alpha)


def front_laplace(x, y, z, t, dim):
    x2 = x**2
    y2 = y**2
    z2 = z**2
    t2 = t**2
    a2 = alpha**2
    a4 = a2**2
    r = sqrt(x2 + y2 + z2)
    ret = (-exp(- 0.5 * ((r - t)/alpha)**2) / a4 *
           (-2.*t*(x2 + y2 + z2 - (dim-1)*a2/2.)/r + x2 + y2 + z2 + t2 - dim*a2))
    return ret


def ball_laplace(x, y, z, t, dim):
    x2 = x**2
    y2 = y**2
    z2 = z**2
    t2 = t**2
    a2 = alpha**2
    a4 = a2**2
    r = sqrt(x2 + y2 + z2)
    den = (-a2 - 4.*(x2 + y2 + z2 - 2*r*t) - t2)**2

    if dim == 2:
        return 2*alpha*(a2 + t2 - 4*x2 - 4*y2)/r/den
    return 4*alpha*(a2 + t2 - 4.*r*t)/r/den


def composite_func(pt, t, f0, f1):
    dim = len(pt)
    x = pt[0]
    y = pt[1]
    z = 0.0
    if dim == 3:
        z = pt[2]
    if (problem == 0):
        if (nfeatures <= 1):
            return f0(x, y, z, t, dim)
        else:
            i = np.arange(nfeatures)
            x0 = 0.5*cos(2 * pi * i / nfeatures)
            y0 = 0.5*sin(2 * pi * i / nfeatures)
            return np.sum(f0(x - x0, y - y0, z, t, dim))
    else:
        i = np.arange(nfeatures)
        x0 = 0.5*cos(2 * pi * i / nfeatures + pi*t)
        y0 = 0.5*sin(2 * pi * i / nfeatures + pi*t)
        return np.sum(f1(x - x0, y - y0, z, 0.25, dim))


class BdrCoefficient(mfem.PyCoefficientT):
    def EvalValue(self, pt, t):
        return composite_func(pt, t, front, ball)


class RhsCoefficient(mfem.PyCoefficientT):
    def EvalValue(self, pt, t):
        # print 'rhs', composite_func(pt, t, front_laplace, ball_laplace)
        return composite_func(pt, t, front_laplace, ball_laplace)


def UpdateAndRebalance(pmesh, fespace, x, a, b):
    # Update the space: recalculate the number of DOFs and construct a matrix
    # that will adjust any GridFunctions to the new mesh state.
    fespace.Update()

    # Interpolate the solution on the new mesh by applying the transformation
    # matrix computed in the finite element space. Multiple GridFunctions could
    # be updated here.
    x.Update()

    if pmesh.Nonconforming():
        # Load balance the mesh.
        pmesh.Rebalance()
        # Update the space again, this time a GridFunction redistribution matrix
        # is created. Apply it to the solution.
        fespace.Update()
        x.Update()

    # Inform the linear and bilinear forms that the space has changed.
    a.Update()
    b.Update()

    # Free any transformation matrices to save memory.
    fespace.UpdatesFinished()

# 3. Read the (serial) mesh from the given mesh file on all processors.  We
#    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
#    and volume meshes with the same code.


meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
sdim = mesh.SpaceDimension()

# 4. Project a NURBS mesh to a piecewise-quadratic curved mesh. Make sure
#    that the mesh is non-conforming if it has quads or hexes and refine it
if (mesh.NURBSext):
    mesh.UniformRefinement()
    if ref_levels > 0:
        ref_levels = ref_levels-1
    mesh.SetCurvature(2)

mesh.EnsureNCMesh()
for l in range(ref_levels):
    mesh.UniformRefinement()

# 5. Define a parallel mesh by partitioning the serial mesh.  Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
ess_bdr.Assign(1)

# 6. Define a finite element space on the mesh. The polynomial order is one
#   (linear) by default, but this can be changed on the command line.
fec = mfem.H1_FECollection(order, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fec)

# 7. As in Example 1p, we set up bilinear and linear forms corresponding to
#    the Laplace problem -\Delta u = 1. We don't assemble the discrete
#    problem yet, this will be done in the inner loop.
a = mfem.ParBilinearForm(fespace)
b = mfem.ParLinearForm(fespace)

one = mfem.ConstantCoefficient(1.0)
bdr = BdrCoefficient()
rhs = RhsCoefficient()

integ = mfem.DiffusionIntegrator(one)
a.AddDomainIntegrator(integ)
b.AddDomainIntegrator(mfem.DomainLFIntegrator(rhs))

# 8. The solution vector x and the associated finite element grid function
#    will be maintained over the AMR iterations.
x = mfem.ParGridFunction(fespace)

# 9. Connect to GLVis.
if visualization:
    sout = mfem.socketstream("localhost", 19916)
    sout.precision(8)

# 10. As in Example 6p, we set up a Zienkiewicz-Zhu estimator that will be
#     used to obtain element error indicators. The integrator needs to
#     provide the method ComputeElementFlux. We supply an L2 space for the
#     discontinuous flux and an H(div) space for the smoothed flux.
flux_fec = mfem.L2_FECollection(order, dim)
flux_fes = mfem.ParFiniteElementSpace(pmesh, flux_fec, sdim)
smooth_flux_fec = mfem.RT_FECollection(order-1, dim)
smooth_flux_fes = mfem.ParFiniteElementSpace(pmesh, smooth_flux_fec)
estimator = mfem.L2ZienkiewiczZhuEstimator(integ, x, flux_fes,
                                           smooth_flux_fes)

# 11. As in Example 6p, we also need a refiner. This time the refinement
#     strategy is based on a fixed threshold that is applied locally to each
#     element. The global threshold is turned off by setting the total error
#     fraction to zero. We also enforce a maximum refinement ratio between
#     adjacent elements.
refiner = mfem.ThresholdRefiner(estimator)
refiner.SetTotalErrorFraction(0.0)   # purely local threshold
refiner.SetLocalErrorGoal(max_elem_error)
refiner.PreferConformingRefinement()
refiner.SetNCLimit(nc_limit)

# 12. A derefiner selects groups of elements that can be coarsened to form
#     a larger element. A conservative enough threshold needs to be set to
#      prevent derefining elements that would immediately be refined again.
derefiner = mfem.ThresholdDerefiner(estimator)
derefiner.SetThreshold(hysteresis * max_elem_error)
derefiner.SetNCLimit(nc_limit)

# 13. The outer time loop. In each iteration we update the right hand side,
#     solve the problem on the current mesh, visualize the solution and
#     refine the mesh as many times as necessary. Then we derefine any
#     elements which have very small errors.
# x.Assign(0.0)
time = 0.0
while (time < t_final + 1e-10):
    if myid == 0:
        print("Time " + str(time) + "\n\nRefinement:")
    bdr.SetTime(time)
    rhs.SetTime(time)

    # Make sure errors will be recomputed in the following.
    refiner.Reset()
    derefiner.Reset()

    # 14. The inner refinement loop. At the end we want to have the current
    #     time step resolved to the prescribed tolerance in each element.
    ref_it = 1
    while(True):
        global_dofs = fespace.GlobalTrueVSize()
        if myid == 0:
            print_txt = ("Iteration: " + str(ref_it) + ", number of unknowns: "
                         + str(global_dofs))

        # 15. Recompute the field on the current mesh: assemble the stiffness
        #     matrix and the right-hand side.
        a.Assemble()
        b.Assemble()

        # 16. Project the exact solution to the essential DOFs.
        x.ProjectBdrCoefficient(bdr, ess_bdr)

        # 17. Create and solve the parallel linear system.
        ess_tdof_list = mfem.intArray()
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

        A = mfem.HypreParMatrix()
        B = mfem.Vector()
        X = mfem.Vector()
        a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)

        amg = mfem.HypreBoomerAMG(A)
        amg.SetPrintLevel(0)
        pcg = mfem.HyprePCG(A)
        pcg.SetTol(1e-12)
        pcg.SetMaxIter(200)
        pcg.SetPrintLevel(0)
        pcg.SetPreconditioner(amg)
        pcg.Mult(B, X)

        # 18. Extract the local solution on each processor.
        a.RecoverFEMSolution(X, b, x)

        # 19. Send the solution by socket to a GLVis server
        if visualization:
            sout.send_text("parallel " + str(num_procs) + " " + str(myid))
            sout.precision(8)
            sout.send_solution(pmesh, x)

        # 20. Apply the refiner on the mesh. The refiner calls the error
        #     estimator to obtain element errors, then it selects elements to
        #     be refined and finally it modifies the mesh. The Stop() method
        #     determines if all elements satisfy the local threshold.
        refiner.Apply(pmesh)
        if myid == 0:
            print(print_txt + ", total error: " +
                  str(estimator.GetTotalError()))
        # 21. Quit the AMR loop if the termination criterion has been met
        if refiner.Stop():
            a.Update()
            break

        # 22. Update the space, interpolate the solution, rebalance the mesh.
        UpdateAndRebalance(pmesh, fespace, x, a, b)
        ref_it = ref_it + 1
    # 21. Use error estimates from the last inner iteration to check for
    #     possible derefinements. The derefiner works similarly as the
    #     refiner. The errors are not recomputed because the mesh did not
    #     change (and also the estimator was not Reset() at this time).
    if derefiner.Apply(pmesh):
        if myid == 0:
            print("Derefined elements.")
        #  22. Update the space and interpolate the solution.
        UpdateAndRebalance(pmesh, fespace, x, a, b)

    time = time + 0.01



'''
   MFEM example 26
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(order=1,
        href=1,
        mesh_file='',
        nd=2,
        initref=1,
        length=1.0,
        k=0.5,
        sol=1,
        visualization=True):

    if nd == 2:
        mesh = mfem.Mesh.MakeCartesian2D(
            1, 1, mfem.Element.QUADRILATERAL, True, length, length, False)
    else:
        mesh = mfem.Mesh.MakeCartesian3D(
            1, 1, 1, mfem.Element.HEXAHEDRON, True, length, length, length)

    sdim = mesh.SpaceDimension()
    dim = mesh.Dimension()

    for i in range(initref):
        mesh.UniformRefinement()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    mesh.Clear()
    pmesh.UniformRefinement()

    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)

    #fespaces = mfem.vector_ParFiniteElementSpace(href+1)
    fespaces = [None]*(href + 1)
    ParMeshes = [None]*(href + 1)
    P = [None] * href  # mfem.vector_HypreParMatrix(href)

    for i in range(href):
        ParMeshes[i] = mfem.ParMesh(pmesh)
        fespaces[i] = mfem.ParFiniteElementSpace(fespace, ParMeshes[i])

        pmesh.UniformRefinement()
        fespace.Update()
        Tr = mfem.OperatorHandle(mfem.Operator.Hypre_ParCSR)
        fespace.GetTrueTransferOperator(fespaces[i], Tr)
        Tr.SetOperatorOwner(False)
        P[i] = Tr.AsHypreParMatrix()

    fespaces[href] = mfem.ParFiniteElementSpace(fespace)

    omega = 2.0 * pi * k

    #
    #  functions (these are JITed using Numba insdie run())
    #

    # local parameters used in JIT function
    from numba import types

    params = {"dim": dim, "omega": omega, "sol": sol}

    sig = types.void(types.CPointer(types.double),
                     types.float64[:],
                     types.float64[:])

    get_sol_Rd = mfem.jit.func(sig, params=params)(get_helmholtz_solution_Re)
    get_sol_Im = mfem.jit.func(sig, params=params)(get_helmholtz_solution_Im)

    params["get_helmholtz_solution_Re"] = get_sol_Rd
    params["get_helmholtz_solution_Im"] = get_sol_Im

    f_Re = mfem.jit.scalar(params=params)(f_exact_Re)
    g_Re = mfem.jit.scalar(params=params)(g_exact_Re)
    grad_Re = mfem.jit.vector(sdim=sdim, params=params)(grad_exact_Re)

    f_Im = mfem.jit.scalar(params=params)(f_exact_Im)
    g_Im = mfem.jit.scalar(params=params)(g_exact_Im)
    grad_Im = mfem.jit.vector(sdim=sdim, params=params)(grad_exact_Im)

    # ParLinearForm *b_Re(new ParLinearForm);
    b = mfem.ParComplexLinearForm(fespace, mfem.ComplexOperator.HERMITIAN)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(f_Re),
                          mfem.DomainLFIntegrator(f_Im))

    if sol >= 0:
        # if exact solution exists. Otherwise use homogeneous impedence
        # (gradp . n + i omega p = 0)
        b.AddBoundaryIntegrator(mfem.BoundaryNormalLFIntegrator(grad_Re),
                                mfem.BoundaryNormalLFIntegrator(grad_Im))
        b.AddBoundaryIntegrator(mfem.BoundaryLFIntegrator(g_Re),
                                mfem.BoundaryLFIntegrator(g_Im))
    b.real().Assign(0.0)
    b.imag().Assign(0.0)

    b.Assemble()

    #  7. Set up the bilinear form (Real and Imaginary part)
    one = mfem.ConstantCoefficient(1.0)
    zero = mfem.ConstantCoefficient(0.0)
    neg_omega = mfem.ConstantCoefficient(-omega**2)

    a = mfem.ParSesquilinearForm(fespace, mfem.ComplexOperator.HERMITIAN)
    impedance = mfem.ConstantCoefficient(omega)

    bdr_attr = mfem.intArray([1]*pmesh.bdr_attributes.Max())
    imp_rest = mfem.RestrictedCoefficient(impedance, bdr_attr)

    a.AddDomainIntegrator(mfem.DiffusionIntegrator(one), None)

    # Just putting imaginary integrator with zero just to keep sparsity pattern of real
    # and imaginary part the same
    a.AddDomainIntegrator(mfem.MassIntegrator(
        neg_omega), mfem.MassIntegrator(zero))
    a.AddBoundaryIntegrator(mfem.BoundaryMassIntegrator(
        zero), mfem.BoundaryMassIntegrator(imp_rest))
    a.Assemble(0)

    ess_tdof_list = mfem.intArray()
    ess_bdr = mfem.intArray([0] * pmesh.bdr_attributes.Max())
    fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    #  Solution grid function
    p_gf = mfem.ParComplexGridFunction(fespace)
    p_gf_ex = mfem.ParComplexGridFunction(fespace)

    p_Re = mfem.jit.scalar(params=params)(p_exact_Re)
    p_Im = mfem.jit.scalar(params=params)(p_exact_Im)

    p_gf.Assign(0.0)
    p_gf_ex.ProjectCoefficient(p_Re, p_Im)
    if sol >= 0:
        p_gf.ProjectBdrCoefficient(p_Re, p_Im, ess_bdr)

    Ah = mfem.OperatorHandle()
    B = mfem.Vector()
    X = mfem.Vector()

    a.FormLinearSystem(ess_tdof_list, p_gf, b, Ah, X, B)

    AZ = Ah.AsComplexHypreParMatrix()

    # ComplexMGSolver * M = new ComplexMGSolver(AZ,P,fespaces);
    # M->SetTheta(0.25);

    M = mfem.ComplexSchwarzSmoother(fespace.GetParMesh(), 0, fespace, AZ)

    gmres = mfem.GMRESSolver(MPI.COMM_WORLD)
    gmres.SetPrintLevel(1)
    gmres.SetMaxIter(2000)
    gmres.SetKDim(200)
    gmres.SetRelTol(1e-12)
    gmres.SetAbsTol(0.0)
    gmres.SetOperator(AZ)
    gmres.SetPreconditioner(M)
    gmres.Mult(B, X)

    a.RecoverFEMSolution(X, B, p_gf)

    if sol >= 0:
        h1_norm_type = 1

        L2err_Re = p_gf.real().ComputeL2Error(p_Re)
        L2err_Im = p_gf.imag().ComputeL2Error(p_Im)

        loc_H1err_Re = p_gf.real().ComputeH1Error(
            p_Re, grad_Re, one, 1.0, h1_norm_type)
        loc_H1err_Im = p_gf.imag().ComputeH1Error(
            p_Im, grad_Im, one, 1.0, h1_norm_type)

        H1err_Re = mfem.GlobalLpNorm(2.0, loc_H1err_Re, MPI.COMM_WORLD)
        H1err_Im = mfem.GlobalLpNorm(2.0, loc_H1err_Im, MPI.COMM_WORLD)

        L2error = sqrt(L2err_Re*L2err_Re + L2err_Im*L2err_Im)
        H1error = sqrt(H1err_Re*H1err_Re + H1err_Im*H1err_Im)
        if myid == 0:
            print(" || p_h - p ||_{H^1} = " + "{:g}".format(H1error))
            print(" || p_h - p ||_{L^1} = " + "{:g}".format(L2error))

    if visualization:
        keys = "keys macF\n" if dim == 3 else "keys amrRljcUUuu\n"
        if dim == 2:
            keys = "keys mrRljc\n"
        else:
            keys = "keys mc\n"

        sol_sock_re = mfem.socketstream("localhost", 19916)
        sol_sock_re.precision(8)
        sol_sock_re << "parallel " << num_procs << " " << myid << "\n"
        sol_sock_re << "solution\n" << pmesh << p_gf.real() << keys
        sol_sock_re << "window_title 'Numerical Pressure (real part)' "
        sol_sock_re.flush()

        sol_sock_im = mfem.socketstream("localhost", 19916)
        sol_sock_im.precision(8)
        sol_sock_im << "parallel " << num_procs << " " << myid << "\n"
        sol_sock_im << "solution\n" << pmesh << p_gf.imag() << keys
        sol_sock_re << "window_title 'Numerical Pressure (imag part)' "
        sol_sock_im.flush()


def get_helmholtz_solution_Re(x, dp, p_d2p):
    if sol == 0:   # polynomial
        if dim == 3:
            p = x[0]*(1.0 - x[0]) * x[1]*(1.0 - x[1]) * x[2]*(1.0 - x[2])
            dp[0] = (1.0 - 2.0 * x[0]) * x[1]*(1.0 - x[1]) * x[2]*(1.0 - x[2])
            dp[1] = (1.0 - 2.0 * x[1]) * x[0]*(1.0 - x[0]) * x[2]*(1.0 - x[2])
            dp[2] = (1.0 - 2.0 * x[2]) * x[0]*(1.0 - x[0]) * x[1]*(1.0 - x[1])
            d2p = (-2.0*(-1.0 + x[0]) * x[0] * (-1.0 + x[1]) * x[1]
                   - 2.0*(-1.0 + x[0]) * x[0] * (-1.0 + x[2]) * x[2]
                   - 2.0*(-1.0 + x[1]) * x[1] * (-1.0 + x[2]) * x[2])
        else:
            p = x[1] * (1.0 - x[1]) * x[0] * (1.0 - x[0])
            dp[0] = (1.0 - 2.0 * x[0]) * x[1]*(1.0 - x[1])
            dp[1] = (1.0 - 2.0 * x[1]) * x[0]*(1.0 - x[0])
            d2p = (- 2.0 * x[1] * (1.0 - x[1])
                   - 2.0 * x[0] * (1.0 - x[0]))

    elif sol == 1:  # Plane wave
        if dim == 2:
            alpha = omega/sqrt(2)
            p = cos(alpha * (x[0] + x[1]))
            dp[0] = -alpha * sin(alpha * (x[0] + x[1]))
            dp[1] = dp[0]
            d2p = -2.0 * alpha * alpha * p
        else:
            alpha = omega/sqrt(3)
            p = cos(alpha * (x[0] + x[1] + x[2]))
            dp[0] = -alpha * sin(alpha * (x[0] + x[1] + x[2]))
            dp[1] = dp[0]
            dp[2] = dp[0]
            d2p = -3.0 * alpha * alpha * p

    elif sol == 2:
        if dim == 2:
            # shift to avoid singularity
            shift = 0.1
            x0 = x[0] + shift
            x1 = x[1] + shift

            r = sqrt(x0 * x0 + x1 * x1)

            p = cos(omega * r)

            r_x = x0 / r
            r_y = x1 / r
            r_xx = (1.0 / r) * (1.0 - r_x * r_x)
            r_yy = (1.0 / r) * (1.0 - r_y * r_y)

            dp[0] = - omega * sin(omega * r) * r_x
            dp[1] = - omega * sin(omega * r) * r_y

            d2p = (-omega*omega * cos(omega * r)*r_x * r_x - omega * sin(omega*r) * r_xx
                   - omega*omega * cos(omega * r)*r_y * r_y - omega * sin(omega*r) * r_yy)
        else:
            # shift to avoid singularity
            shift = 0.1
            x0 = x[0] + shift
            x1 = x[1] + shift
            x2 = x[2] + shift

            r = sqrt(x0 * x0 + x1 * x1 + x2 * x2)

            p = cos(omega * r)

            r_x = x0 / r
            r_y = x1 / r
            r_z = x2 / r
            r_xx = (1.0 / r) * (1.0 - r_x * r_x)
            r_yy = (1.0 / r) * (1.0 - r_y * r_y)
            r_zz = (1.0 / r) * (1.0 - r_z * r_z)

            dp[0] = - omega * sin(omega * r) * r_x
            dp[1] = - omega * sin(omega * r) * r_y
            dp[2] = - omega * sin(omega * r) * r_z

            d2p = (-omega*omega * cos(omega * r)*r_x * r_x - omega * sin(omega*r) * r_xx
                   - omega*omega * cos(omega * r)*r_y *
                   r_y - omega * sin(omega*r) * r_yy
                   - omega*omega * cos(omega * r)*r_z * r_z - omega * sin(omega*r) * r_zz)

    p_d2p[0] = p
    p_d2p[1] = d2p


def get_helmholtz_solution_Im(x, dp, p_d2p):
    if sol == 0:  # polynomial
        if dim == 3:
            p = x[0]*(1.0 - x[0]) * x[1]*(1.0 - x[1]) * x[2]*(1.0 - x[2])
            dp[0] = (1.0 - 2.0 * x[0]) * x[1]*(1.0 - x[1]) * x[2]*(1.0 - x[2])
            dp[1] = (1.0 - 2.0 * x[1]) * x[0]*(1.0 - x[0]) * x[2]*(1.0 - x[2])
            dp[2] = (1.0 - 2.0 * x[2]) * x[0]*(1.0 - x[0]) * x[1]*(1.0 - x[1])
            d2p = (-2.0*(-1.0 + x[0]) * x[0] * (-1.0 + x[1]) * x[1]
                   - 2.0*(-1.0 + x[0]) * x[0] * (-1.0 + x[2]) * x[2]
                   - 2.0*(-1.0 + x[1]) * x[1] * (-1.0 + x[2]) * x[2])
        else:
            p = x[1] * (1.0 - x[1]) * x[0] * (1.0 - x[0])
            dp[0] = (1.0 - 2.0 * x[0]) * x[1]*(1.0 - x[1])
            dp[1] = (1.0 - 2.0 * x[1]) * x[0]*(1.0 - x[0])
            d2p = (- 2.0 * x[1] * (1.0 - x[1])
                   - 2.0 * x[0] * (1.0 - x[0]))
    elif sol == 1:  # plane wave
        if dim == 2:
            alpha = omega/sqrt(2)
            p = -sin(alpha * (x[0] + x[1]))
            dp[0] = -alpha * cos(alpha * (x[0] + x[1]))
            dp[1] = dp[0]
            d2p = -2.0 * alpha * alpha * p
        else:
            alpha = omega/sqrt(3)
            p = -sin(alpha * (x[0] + x[1] + x[2]))
            dp[0] = -alpha * cos(alpha * (x[0] + x[1] + x[2]))
            dp[1] = dp[0]
            dp[2] = dp[0]
            d2p = -3.0 * alpha * alpha * p
    p_d2p[0] = p
    p_d2p[1] = d2p


def p_exact_Re(x):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Re(x, dp, p_d2p)
    return p_d2p[0]


def p_exact_Im(x):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Im(x, dp, p_d2p)
    return p_d2p[0]


def f_exact_Re(x):
    p_d2p_re = np.empty(2)
    dp_re = np.empty(3)
    get_helmholtz_solution_Re(x, dp_re, p_d2p_re)
    p_re = p_d2p_re[0]
    d2p_re = p_d2p_re[1]
    f_re = -d2p_re - omega * omega * p_re

    return f_re


def f_exact_Im(x):
    p_d2p_im = np.empty(2)
    dp_im = np.empty(3)

    get_helmholtz_solution_Im(x, dp_im, p_d2p_im)
    p_im = p_d2p_im[0]
    d2p_im = p_d2p_im[1]

    f_im = -d2p_im - omega * omega * p_im
    return f_im


def grad_exact_Re(x, ret):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Re(x, dp, p_d2p)
    for i in range(dim):
        ret[i] = dp[i]


def grad_exact_Im(x, ret):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Im(x, dp, p_d2p)
    for i in range(dim):
        ret[i] = dp[i]

# define impedence coefficient: i omega p


def g_exact_Re(x):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Im(x, dp, p_d2p)
    return -omega * p_d2p[0]


def g_exact_Im(x):
    p_d2p = np.empty(2)
    dp = np.empty(3)
    get_helmholtz_solution_Re(x, dp, p_d2p)
    return omega * p_d2p[0]


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='complex MG')

    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Elemet order")
    parser.add_argument('-nd', '--ndim',
                        action='store', default=2, type=int,
                        help='Problem space dimension')
    parser.add_argument("-sol", "--exact_sol",
                        action='store', default=1, type=int,
                        help="Exact solution flag - 0:polynomial, " +
                        "1: plane wave, -1: unknown exact")
    parser.add_argument("-k", "--wavelengths",
                        action='store', default=0.5, type=float,
                        help="Number of wavelengths.")
    parser.add_argument('-length', '--length',
                        action='store', default=1, type=float,
                        help='length of domain')
    parser.add_argument('-href', '--href',
                        action='store', default=1, type=int,
                        help="Number of geometric refinements done " +
                        "prior to order refinements.")
    parser.add_argument('-initref', '--initref',
                        action='store', default=1, type=int,
                        help='initial refinement')
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    mesh_file = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    run(order=args.order,
        href=args.href,
        mesh_file=mesh_file,
        nd=args.ndim,
        initref=args.initref,
        length=args.length,
        k=args.wavelengths,
        sol=args.exact_sol,
        visualization=args.visualization)



'''
   MFEM example 26
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(order_refinements=2,
        geometric_refinements=0,
        mesh_file='',
        visualization=True,
        device='cpu'):

    class DiffusionMultigrid(mfem.PyGeometricMultigrid):
        def __init__(self, fespaces, ess_bdr):
            mfem.PyGeometricMultigrid.__init__(self, fespaces, ess_bdr)

            self.smoothers = []
            self.ConstructCoarseOperatorAndSolver(
                fespaces.GetFESpaceAtLevel(0),)

            for level in range(1, fespaces.GetNumLevels()):
                self.ConstructOperatorAndSmoother(fespaces.GetFESpaceAtLevel(level),
                                                  level)

        def ConstructBilinearForm(self, fespace, partial_assembly):
            form = mfem.ParBilinearForm(fespace)
            if partial_assembly:
                form.SetAssemblyLevel(mfem.AssemblyLevel_PARTIAL)
            form.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
            form.Assemble()
            self.AppendBilinearForm(form)

        def ConstructCoarseOperatorAndSolver(self, coarse_fespace,):

            self.ConstructBilinearForm(coarse_fespace, False)

            hypreCoarseMat = mfem.HypreParMatrix()
            self.bfs[0].FormSystemMatrix(self.essentialTrueDofs[0],
                                         hypreCoarseMat)

            amg = mfem.HypreBoomerAMG(hypreCoarseMat)
            amg.SetPrintLevel(-1)

            pcg = mfem.CGSolver(MPI.COMM_WORLD)
            pcg.SetPrintLevel(-1)
            pcg.SetMaxIter(200)
            pcg.SetRelTol(sqrt(1e-4))
            pcg.SetAbsTol(0.0)
            pcg.SetOperator(hypreCoarseMat)
            pcg.SetPreconditioner(amg)

            self.smoothers.append((pcg, hypreCoarseMat))
            self.AddLevel(hypreCoarseMat, pcg, False, False)
            self.amg = amg

        def ConstructOperatorAndSmoother(self, fespace, level):
            self.ConstructBilinearForm(fespace, True)

            opr = mfem.OperatorPtr()
            opr.SetType(mfem.Operator.ANY_TYPE)

            self.bfs[level].FormSystemMatrix(
                self.essentialTrueDofs[level], opr)
            opr.SetOperatorOwner(False)

            diag = mfem.Vector(fespace.GetTrueVSize())
            self.bfs[level].AssembleDiagonal(diag)

            smoother = mfem.OperatorChebyshevSmoother(opr.Ptr(), diag,
                                                      self.essentialTrueDofs[level],
                                                      2,
                                                      fespace.GetParMesh().GetComm())

            self.smoothers.append((smoother, opr))
            self.AddLevel(opr.Ptr(), smoother, False, False)

    device = mfem.Device(device)
    if myid == 0:
        device.Print()

    mesh = mfem.Mesh(mesh_file, 1, 1)
    dim = mesh.Dimension()

    ref_levels = int(np.floor(np.log(1000. / mesh.GetNE()) / np.log(2.) / dim))
    for i in range(ref_levels):
        mesh.UniformRefinement()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    pmesh.UniformRefinement()
    pmesh.UniformRefinement()

    # 5. Define a finite element space hierarchy on the mesh. Here we use
    #    continuous Lagrange finite elements. We start with order 1 on the
    #    coarse level and geometrically refine the spaces by the specified
    #    amount. Afterwards, we increase the order of the finite elements
    #    by a factor of 2 for each additional level.
    fec = mfem.H1_FECollection(1, dim)
    coarse_fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    # note ownM and ownFES are False since they are managed by Python GC.
    fespaces = mfem.ParFiniteElementSpaceHierarchy(
        pmesh, coarse_fespace, False, False)

    collections = []
    collections.append(fec)
    for level in range(geometric_refinements):
        fespaces.AddUniformlyRefinedLevel()
    for level in range(order_refinements):
        collections.append(mfem.H1_FECollection(2**(level+1), dim))
        fespaces.AddOrderRefinedLevel(collections[-1])

    global_dof = fespaces.GetFinestFESpace().GlobalTrueVSize()
    if myid == 0:
        print("Number of finite element unknowns: " + str(global_dof))

    # 6. Set up the linear form b(.) which corresponds to the right-hand side of
    #    the FEM linear system, which in this case is (1,phi_i) where phi_i are
    #    the basis functions in the finite element fespace.
    b = mfem.ParLinearForm(fespaces.GetFinestFESpace())
    one = mfem.ConstantCoefficient(1.0)
    b.AddDomainIntegrator(mfem.DomainLFIntegrator(one))
    b.Assemble()

    # 7. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x with initial guess of zero,
    #    which satisfies the boundary conditions.
    x = mfem.ParGridFunction(fespaces.GetFinestFESpace())
    x.Assign(0.0)

    # 8. Create the multigrid operator using the previously created
    #    FiniteElementSpaceHierarchy and additional boundary information. This
    #    operator is then used to create the MultigridSolver as a preconditioner
    #    in the iterative solver.
    ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
    ess_bdr.Assign(1)

    M = DiffusionMultigrid(fespaces, ess_bdr)
    M.SetCycleType(mfem.Multigrid.CycleType_VCYCLE, 1, 1)

    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()

    M.FormFineLinearSystem(x, b, A, X, B)
    if myid == 0:
        print("Size of linear system: " + str(A.Height()))

    # 9. Solve the linear system A X = B.
    cg = mfem.CGSolver(MPI.COMM_WORLD)
    cg.SetRelTol(1e-12)
    cg.SetMaxIter(200)
    cg.SetPrintLevel(1)
    cg.SetOperator(A.Ptr())
    cg.SetPreconditioner(M)

    import sys
    sys.setrecursionlimit(200000)
    cg.Mult(B, X)

    # 10. Recover the solution as a finite element grid function.
    M.RecoverFineFEMSolution(X, b, x)

    # 11. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf".
    fespaces.GetFinestFESpace().GetMesh().Print("mesh"+smyid, 8)
    x.Save("sol"+smyid, 8)

    # 12. Send the solution by socket to a GLVis server.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock.precision(8)
        sol_sock << "solution\n" << fespaces.GetFinestFESpace().GetMesh() << x
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex26p (Multigrid Preconditioner)')

    parser.add_argument('-m', '--mesh',
                        default='star.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-gr', '--geometric-refinement',
                        action='store', default=0, type=int,
                        help="Number of geometric refinements done prior to order refinements.")
    parser.add_argument('-or', '--order-refinement',
                        action='store', default=2, type=int,
                        help="Number of order refinements. Finest level in the hierarchy has order 2^{or}.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    mesh_file = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))

    run(order_refinements=args.order_refinement,
        geometric_refinements=args.geometric_refinement,
        mesh_file=mesh_file,
        visualization=args.visualization,
        device=args.device)



# ----------------------------------------------------------------------------------------
# Routines for gslib
# ----------------------------------------------------------------------------------------
import sys
import os
import re
import subprocess

import build_globals as bglb
from build_consts import *
from build_utils import *

__all__ = ['make_gslib']

def make_gslib(serial=False):
    if bglb.verbose:
        print("Building gslib")

    path = os.path.join(extdir, 'gslib')
    if not os.path.exists(path):
        assert False, "gslib is not downloaded"

    pwd = chdir(path)
    make_call(['make', 'clean'])
    if serial:
        command = ['make', 'CC=' + bglb.cc_command, 'MPI=0', 'CFLAGS=-fPIC']
        make_call(command)
        command = ['make', 'MPI=0', 'DESTDIR=' + bglb.gslibs_prefix]
        make_call(command)
    else:
        command = ['make', 'CC=' + bglb.mpicc_command, 'CFLAGS=-O2 -fPIC']
        make_call(command)
        command = ['make', 'DESTDIR=' + bglb.gslibp_prefix]
        make_call(command)
    os.chdir(pwd)



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    
def run_test():
    a = mfem.uintArray(5)
    b = mfem.int8Array(5)
    c = mfem.int64Array(5)
    d = mfem.boolArray(5)

    def check(a, value):
        a.GetDataArray()[:] = value
        print(a[0], a[-1], type(a[0]), )
        assert a[0] == value, "Array data element does not agree"


    check(a, 5)
    check(b, 100)
    check(c, 1000)
    check(d, True)    
    
    
if __name__=='__main__':
    run_test()



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0

def InArray(T, sz, i):
    hoge = mfem.intArray((T, sz))
    ret = (i in hoge.ToList())
    return ret
    
def IndicesAreConnected(e2e, i, j):
    return (InArray(e2e.GetRow(i), e2e.RowSize(i), j) and
            InArray(e2e.GetRow(j), e2e.RowSize(j), i))

def run_test():
    n = 3

    ### 1D Periodic mesh
    orig_mesh = mfem.Mesh.MakeCartesian1D(n)
    translations = (mfem.Vector([1.0]),)
    #vv =mfem.vector_Vector(translations)
    mapping = orig_mesh.CreatePeriodicVertexMapping(translations)

    mesh = mfem.Mesh.MakePeriodic(orig_mesh, mapping)

    assert mesh.GetNV() == n, "GetNV is not correct"

    e2e = mesh.ElementToElementTable()
    assert IndicesAreConnected(e2e, 0, 2), "wrong connection"
    assert IndicesAreConnected(e2e, 0, 1), "wrong connection"
    assert IndicesAreConnected(e2e, 1, 2), "wrong connection"

    ### 2D Periodic mesh
    els = (mfem.Element.TRIANGLE, mfem.Element.QUADRILATERAL)

    for el in els:
        orig_mesh = mfem.Mesh.MakeCartesian2D(n, n, el, False, 1, 1, False)
        translations = (mfem.Vector([1., 0.]),
                        mfem.Vector([0., 1.]),)    
        mapping = orig_mesh.CreatePeriodicVertexMapping(translations)
        mesh = mfem.Mesh.MakePeriodic(orig_mesh, mapping)

        assert mesh.GetNV() == (n-1)**2 + 2*(n-1) + 1, "GetNV is not correct"

        if el == mfem.Element.QUADRILATERAL:
            e2e = mesh.ElementToElementTable()
            for i in range(n):
                assert IndicesAreConnected(e2e, i, i + n*(n-1)), "wrong connection"
                assert IndicesAreConnected(e2e, i*n, n-1 + i*n), "wrong connection"

    ### 3D Periodic mesh
    els = (mfem.Element.TETRAHEDRON,
           mfem.Element.HEXAHEDRON,
           mfem.Element.WEDGE)
    for el in els:
        orig_mesh = mfem.Mesh.MakeCartesian3D(n, n, n, el, 1, 1, 1, False)
        translations = (mfem.Vector([1., 0., 0.]),
                        mfem.Vector([0., 1., 0.]),
                        mfem.Vector([0., 0., 1.]),)
        mapping = orig_mesh.CreatePeriodicVertexMapping(translations)
        mesh = mfem.Mesh.MakePeriodic(orig_mesh, mapping)

        assert mesh.GetNV() == (n-1)**3 + 3*(n-1)**2 + 3*(n-1) +1, "GetNV is not correct"        
        if el == mfem.Element.HEXAHEDRON:
            e2e = mesh.ElementToElementTable()
            n2 = n*n
            for j in range(n):
                for i in range(n):       
                    assert IndicesAreConnected(e2e,
                                               i + j*n,
                                               i + j*n + n2*(n-1)), "wrong connection"
                    assert IndicesAreConnected(e2e,
                                               i + j*n2,
                                               i + j*n2 + n*(n-1)), "wrong connection"
                    assert IndicesAreConnected(e2e,
                                               i*n + j*n2,
                                               i*n + j*n2 + n-1), "wrong connection"



if __name__=='__main__':
    run_test()


